################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import Component

class Priority(Component):
    """
    Raise/lower workgroup priority.
    """
    pass

class ConstantPriority(Priority):
    """
    Priority implementation which does nothing.
    """
    kernel = {"AggressivePerfMode": False}

    def __call__(self, writer, prio, message=""):
        return ""

class AggressivePriority(Priority):
    """
    Priority implementation which does set the priority.

    Keeps track of the previous value in the instance and only sets priority
    if the new priority is different.
    """
    kernel = {"AggressivePerfMode": True}

    def __init__(self, currentPrio=None):
        self.currentPrio = currentPrio

    def __call__(self, writer, prio, message=""):
        if prio == self.currentPrio:
            return ""

        self.currentPrio = prio
        vars = {"endLine": writer.endLine, "prio": prio, "message": message}

        return "s_setprio {prio} // {message}{endLine}".format_map(vars)



################################################################################
#
# Copyright (C) 2020-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import Signature
from ..Common import globalParameters, getCOVFromParam, gfxName

from math import ceil

def getSrcValueType(kernel, isTypeA):
    # special cases for F8 datatypes
    if kernel["ProblemType"]["DataType"].isFloat8():
        srcValueType = "FP8"
    elif kernel["ProblemType"]["DataType"].isBFloat8():
        srcValueType = "BF8"
    elif kernel["ProblemType"]["DataType"].isFloat8BFloat8():
        srcValueType = "FP8" if isTypeA else "BF8"
    elif kernel["ProblemType"]["DataType"].isBFloat8Float8():
        srcValueType = "BF8" if isTypeA else "FP8"
    else:
        srcValueType = kernel["ProblemType"]["DataType"].toNameAbbrev().upper()

    srcValueType = srcValueType.lower()
    return srcValueType

def getDstValueType(kernel):
    # special cases for F8 datatypes
    if kernel["ProblemType"]["DataType"].isFloat8():
        dstValueType = "FP8"
    elif kernel["ProblemType"]["DataType"].isBFloat8():
        dstValueType = "BF8"
    else:
        dstValueType = kernel["ProblemType"]["DataType"].toNameAbbrev().upper()

    dstValueType = dstValueType.lower()
    return dstValueType


# Creates kernel header, compatible with code object version 4 and up. V2 and V3 no longer supported.
class SignatureDefault(Signature):

    # Formats an argument to add to the header
    def addArgument(self, name, size, offset, valueKind, valueType = None, AddrSpaceQual = None):
        self.offset += size
        if valueType is None:
            valueType = f"u{size*8}"
        kStr = ""
        kStr += "      - .name:            %s\n" % name
        kStr += "        .size:            %s\n" % size
        kStr += "        .offset:          %s\n" % offset
        kStr += "        .value_kind:      %s\n" % valueKind
        kStr += "        .value_type:      %s\n" % valueType
        if AddrSpaceQual != None:
            kStr += "        .address_space:   %s\n" % AddrSpaceQual
        return kStr

    def __call__(self, writer):
        kernel = writer.kernel

        kStr = self.commentHeader()

        # begin kernel descriptor
        kStr += ".amdgcn_target \"amdgcn-amd-amdhsa--%s\"%s" \
            % (gfxName(writer.version), writer.endLine)

        kStr += ".text%s" % writer.endLine
        kStr += ".protected %s%s" % (writer.kernelName, writer.endLine)
        kStr += ".globl %s%s" % (writer.kernelName, writer.endLine)
        kStr += ".p2align 8%s" % writer.endLine
        kStr += ".type %s,@function%s" % (writer.kernelName, writer.endLine)

        kStr += ".section .rodata,#alloc%s" % writer.endLine
        kStr += ".p2align 6%s" % writer.endLine
        tWord = "amdhsa_kernel"
        kStr += ".%s %s%s" % (tWord, writer.kernelName, writer.endLine)
        tWord = ".amdhsa_user_sgpr_kernarg_segment_ptr"
        kStr += "  %s 1%s" % (tWord, writer.endLine)

        supportsPreload = globalParameters["AsmCaps"][writer.version]["KernargPreloading"]

        numPreloaded = 0
        numUserSgprs = 2

        if kernel["PreloadKernelArguments"] == 1:
            numPreloaded = writer.sgprPool.numPreloadSGPRs
            numUserSgprs += numPreloaded

        if supportsPreload:
            kStr += f"  .amdhsa_user_sgpr_kernarg_preload_offset 0{writer.endLine}"
            kStr += f"  .amdhsa_user_sgpr_kernarg_preload_length {numPreloaded}{writer.endLine}"

        kStr += f"  .amdhsa_user_sgpr_count {numUserSgprs}{writer.endLine}"

        # kern arg size
        kernArgReg = 0
        kernArgReg += 3*writer.rpga
        kernArgReg += max(1,int(writer.bpeAB/4)) # alpha
        if kernel["ProblemType"]["UseBeta"]:
            kernArgReg += max(1,int(writer.bpeCexternal/4)) # beta
        kernArgReg += kernel["ProblemType"]["NumIndicesC"] # strides
        kernArgReg += kernel["ProblemType"]["NumIndicesC"] # strides
        kernArgReg += len(kernel["ProblemType"]["IndexAssignmentsA"]) # strides
        kernArgReg += len(kernel["ProblemType"]["IndexAssignmentsB"]) # strides
        if not kernel["ProblemType"]["UseInitialStridesAB"]:
            kernArgReg -= 2 # strides
        if not kernel["ProblemType"]["UseInitialStridesCD"]:
            kernArgReg -= 2 # strides
        kernArgReg += kernel["ProblemType"]["NumIndicesSummation"]
        kernArgReg += kernel["ProblemType"]["NumIndicesC"]
        if globalParameters["DebugKernel"]:
            kernArgReg += writer.rpga # debug buffer
        # kernArgBytes = kernArgReg * 4 # bytes/reg

        # register allocation
        totalVgprs = writer.vgprPool.size()
        totalSgprs = writer.sgprPool.size()

        # accumulator offset for Unified Register Files
        vgprCount = totalVgprs
        if writer.archCaps["ArchAccUnifiedRegs"]:
            agprStart = ceil(totalVgprs/8)*8
            vgprCount = agprStart + writer.agprPool.size()

            tWord = ".amdhsa_accum_offset"
            kStr += "  %s %u // accvgpr offset%s" % (tWord, agprStart, writer.endLine)

        tWord = ".amdhsa_next_free_vgpr"
        kStr += "  %s %u // vgprs%s" % (tWord, vgprCount, writer.endLine)
        tWord = ".amdhsa_next_free_sgpr"
        kStr += "  %s %u // sgprs%s" % (tWord, totalSgprs, writer.endLine)

        tWord = ".amdhsa_group_segment_fixed_size"
        if kernel["AggressivePerfMode"]>=2 and kernel["ProblemType"]["DataType"].isDouble() and \
            kernel["ThreadTile0"] == 4 and kernel["ThreadTile1"] == 4 and kernel["WorkGroup"] == [16,16,1]:
            group_segment_size = 32768 # Pad LDS to ensure we run exactly two waves
        else:
            group_segment_size = kernel["LdsNumElements"] * writer.bpeAB
        kStr += "  %s %u // lds bytes%s" % ( tWord, group_segment_size, writer.endLine )

        if writer.archCaps["HasWave32"]:
            if kernel["WavefrontSize"] == 32:
                kStr += "  .amdhsa_wavefront_size32 1 // 32-thread wavefronts%s" % writer.endLine
            else:
                kStr += "  .amdhsa_wavefront_size32 0 // 64-thread wavefronts%s" % writer.endLine

        # other
        kStr += "  .amdhsa_private_segment_fixed_size 0%s" % writer.endLine
        kStr += "  .amdhsa_system_sgpr_workgroup_id_x 1%s" % writer.endLine
        kStr += "  .amdhsa_system_sgpr_workgroup_id_y 1%s" % writer.endLine
        kStr += "  .amdhsa_system_sgpr_workgroup_id_z %u%s" % (1 if kernel["ProblemType"]["NumIndicesC"] > 2 else 0, writer.endLine)
        kStr += "  .amdhsa_system_vgpr_workitem_id 0%s" % writer.endLine
        kStr += "  .amdhsa_float_denorm_mode_32 3%s" % writer.endLine
        kStr += "  .amdhsa_float_denorm_mode_16_64 3%s" % writer.endLine
        kStr += ".end_amdhsa_kernel%s" % writer.endLine
        kStr += ".text%s" % writer.endLine

        kStr += writer.comment3("Optimizations and Config:")
        kStr += writer.comment1("ThreadTile= %u x %u" % (kernel["ThreadTile0"], kernel["ThreadTile1"]))
        kStr += writer.comment1("SubGroup= %u x %u" % (kernel["SubGroup0"], kernel["SubGroup1"]))
        if kernel["EnableMatrixInstruction"]:
          kStr += writer.comment1("VectorWidthA=%u" % (kernel["VectorWidthA"]))
          kStr += writer.comment1("VectorWidthB=%u" % (kernel["VectorWidthB"]))
        else:
          kStr += writer.comment1("VectorWidth=%u" % (kernel["VectorWidth"]))
        glvw = kernel["GlobalLoadVectorWidthA"]
        glvwAStr = "GlobalLoadVectorWidthA=%f"%glvw if glvw < 1 else "GlobalLoadVectorWidthA=%u"%glvw
        glvw = kernel["GlobalLoadVectorWidthB"]
        glvwBStr = "GlobalLoadVectorWidthB=%f"%glvw if glvw < 1 else "GlobalLoadVectorWidthB=%u"%glvw
        kStr += writer.comment1("%s, %s" % (glvwAStr, glvwBStr))
        kStr += writer.comment1("DirectToLdsA=%s" % kernel["DirectToLdsA"])
        kStr += writer.comment1("DirectToLdsB=%s" % kernel["DirectToLdsB"])
        kStr += writer.comment1("UseSgprForGRO=%s" % kernel["_UseSgprForGRO"])

        kStr += ".amdgpu_metadata\n"
        kStr += "---\n"
        kStr += "amdhsa.version:\n"
        kStr += "  - 1\n"
        cov = getCOVFromParam(kernel["CodeObjectVersion"])
        if cov == 4:
            kStr += "  - 1\n"
        elif cov == 5:
            kStr += "  - 2\n"
        kStr += "amdhsa.target: amdgcn-amd-amdhsa--%s%s" \
            % (gfxName(writer.version), writer.endLine)
        kStr += "amdhsa.kernels:\n"
        kStr += "  - .name: %s%s" % (writer.kernelName, writer.endLine)
        kStr += "    .symbol: '%s.kd'%s" % (writer.kernelName, writer.endLine)
        kStr += "    .language:                   %s%s" % ("OpenCL C", writer.endLine)
        kStr += "    .language_version:%s" % writer.endLine
        kStr += "      - 2%s" % writer.endLine
        kStr += "      - 0%s" % writer.endLine
        kStr += "    .args:%s" % writer.endLine
        self.offset = 0

        for arg in writer.sgprPool.kernargs:
            kStr += self.addArgument(name=arg['name'], size=arg['size']*4, offset=self.offset, valueKind="by_value")

        kStr += "    .group_segment_fixed_size:   %u%s" % ( group_segment_size, writer.endLine ) #XXXXXX
        kStr += "    .kernarg_segment_align:      %u%s" % ( 8, writer.endLine )
        kStr += "    .kernarg_segment_size:       %u%s" % (((self.offset+7)//8)*8, writer.endLine) # round up to .kernarg_segment_align
        kStr += "    .max_flat_workgroup_size:    %u%s" % ( kernel["SubGroup0"] * kernel["SubGroup1"] * kernel["LocalSplitU"], writer.endLine )
        kStr += "    .private_segment_fixed_size: %u%s" % ( 0, writer.endLine )
        kStr += "    .sgpr_count:                 %u%s" % ( totalSgprs, writer.endLine )
        kStr += "    .sgpr_spill_count:           %u%s" % ( 0, writer.endLine )
        kStr += "    .vgpr_count:                 %u%s" % ( totalVgprs, writer.endLine )
        kStr += "    .vgpr_spill_count:           %u%s" % ( 0, writer.endLine )
        kStr += "    .wavefront_size:             %u%s" % ( kernel["WavefrontSize"], writer.endLine )

        kStr += "...\n"

        kStr += ".end_amdgpu_metadata\n"

        kStr += "%s:%s" % (writer.kernelName, writer.endLine)

        return kStr



################################################################################
#
# Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from . import Common

import itertools
import os

from pathlib import Path
from typing import List

class Namespace:
    def __init__(self, parent, name=None):
        self.parent = parent
        self.name = name

    def __enter__(self):
        if self.name is None:
            self.parent.write("namespace {")
        else:
            self.parent.write("namespace {} {{".format(self.name))
        self.parent.indent()
        return self

    def __exit__(self, *args, **kwargs):
        self.parent.dedent()
        if self.name is None:
            self.parent.write("} // anonymous namespace")
        else:
            self.parent.write("}} // namespace {}".format(self.name))

class Indent:
    def __init__(self, parent):
        self.parent = parent

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.parent.dedent()

class EmbeddedDataFile:
    def __init__(self, filename, file=None, indent_spaces=4):
        self.filename = filename

        self._indent_spaces = indent_spaces
        self._indent_levels = [0]
        self._open_blocks = []

        self.file = file
        if self.file is None:
            self.file = open(filename, 'w')

        self.write_header()

    def __enter__(self):
        self.file.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        while len(self._open_blocks) > 0:
            b = self._open_blocks.pop()
            b.__exit__(*args, **kwargs)

        self.write_footer()

        self.file.__exit__(*args, **kwargs)

    def namespace(self, name=None):
        ns = Namespace(self, name).__enter__()
        self._open_blocks.append(ns)

    def end_namespace(self, name=None):
        ns = self._open_blocks.pop()
        if not isinstance(ns, Namespace):
            raise RuntimeError("Mismatched block types: expected Namespace, found {}".format(ns.__class__))
        if name != ns.name:
            raise RuntimeError("Mismatched namespace open/close: expected {}, found {}".format(name, ns.name))

        ns.__exit__(None, None, None)

    def write_header(self):
        self.write(Common.CHeader)

        self.write(self.includes)

        self.namespace("Tensile")

    @property
    def include_guard(self):
        return "#pragma once"

    @property
    def includes(self):
        return """
        #include <Tensile/EmbeddedData.hpp>

        #include <Tensile/Contractions.hpp>
        #include <Tensile/Tensile.hpp>
    """

    def get_lines(self, item):
        if isinstance(item, str):
            return item.split('\n')

        if hasattr(item, '__iter__'):
            return item

        return str(item).split('\n')

    def format(self, item):
        out_lines = []
        for line in self.get_lines(item):
            out_lines.append(self.apply_indent(line))

        return '\n'.join(out_lines) + '\n'

    @property
    def indent_level(self):
        if len(self._indent_levels) == 0:
            return 0

        return self._indent_levels[-1]

    def apply_indent(self, line=None):
        if line is None:
            return ' ' * self.indent_level

        line = line.strip()

        if line.startswith('#'):
            return line

        return (' ' * self.indent_level) + line

    def indent(self, spaces=None):
        if spaces is None:
            spaces = self._indent_spaces
        self._indent_levels.append(spaces + self.indent_level)
        return Indent(self)

    def dedent(self):
        self._indent_levels.pop()

    def write(self, *items):
        for item in items:
            self.file.write(self.format(item))

    def comment(self, text):
        self.write(['// ' + line for line in text.split('\n')])

    def write_footer(self):
        self.write('')

    def embed_data(self, assocType, data, nullTerminated=False, comment=None, key=None):
        if nullTerminated:
            empty = False
            data = itertools.chain(data, [0])
        else:
            empty = len(data) == 0
            data = iter(data)

        with Namespace(self):
            if comment is not None:
                self.comment(comment)
            if empty:
                if key is None:
                    self.write("EmbedData<{0}> TENSILE_EMBED_SYMBOL_NAME{{}};".format(assocType))
                else:
                    self.write('EmbedData<{0}> TENSILE_EMBED_SYMBOL_NAME("{1}", {{}});'.format(assocType, key))
                return

            hex_format = '{:#04x}'

            if key is None:
                self.write("EmbedData<{0}> TENSILE_EMBED_SYMBOL_NAME({{".format(assocType))
            else:
                self.write('EmbedData<{0}> TENSILE_EMBED_SYMBOL_NAME("{1}", {{'.format(assocType, key))
            with self.indent():
                line = hex_format.format(next(data))
                for byteIdx, byte in enumerate(data):
                    if byteIdx % 16 == 15:
                        self.write(line + ",")
                        line = hex_format.format(byte)
                    else:
                        line += ', ' + hex_format.format(byte)

                self.write(line + '});')

    def embed_file(self, assocType, filename, nullTerminated=False, key=None):
        with open(filename, 'rb') as f:
          byteArray = bytearray(f.read())
        self.embed_data(assocType, byteArray, nullTerminated, os.path.basename(filename), key)


def generateLibrary(embedFileName: Path, embedLibraryKey: str, masterFile: Path, cppBaseClass: str, codeObjectFiles: List[str]) -> None:
    """Creates embed library source files.

       The goal of an embed library is to convert the master file and code object files
       into a byte array and inject the byte arrays into a C++ source file. This precludes
       the need to load data from the master file and code object files at runtime.

       Args:
           embedFileName: Name of the generated C++ embed source file without a file extension.
           embedLibraryKey: Name or key associated with the embed library.
           masterFile: Path to the master library file (.dat or .yaml).
           cppBaseClass: Name of type used when specializing the C++ EmbeddedData class template.
           codeObjectFiles: List of code object files created by TensileCreateLibrary.
    """
    with EmbeddedDataFile(embedFileName.with_suffix(".temp")) as embedFile:
        embedFile.embed_file(cppBaseClass, masterFile, nullTerminated=True, key=embedLibraryKey)
        for co in codeObjectFiles:
            embedFile.embed_file("SolutionAdapter", co, nullTerminated=False, key=embedLibraryKey)
    os.rename(embedFileName.with_suffix(".temp"), embedFileName.with_suffix(".cpp"))



################################################################################
#
# Copyright (C) 2016-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from copy import deepcopy
import itertools
from .Common import tPrint, printExit, \
        defaultBenchmarkCommonParameters, validParameters, globalParameters, \
        defaultBatchedBenchmarkFinalProblemSizes, defaultBenchmarkFinalProblemSizes
from .CustomKernels import getAllCustomKernelNames
from .SolutionStructs import ProblemType, ProblemSizes


def checkParametersAreValid(param, validParams):
    """Ensures paramaters in params exist and have valid values as specified by validParames"""
    (name, values) = param
    if name == "ProblemSizes":
        return

    if name not in validParams:
        printExit("Invalid parameter name: {}\nValid parameters are {}." \
                .format(name, sorted(validParameters.keys())))

    for value in values:
        if validParams[name] != -1 and value not in validParams[name]:
            msgBase = "Invalid parameter value: {} = {}\nValid values for {} are {}{}."
            msgExt = " (only first 32 combos printed)\nRefer to Common.py for more info" \
                    if len(validParams[name])>32 else ""
            printExit(msgBase.format(name, value, name, validParams[name][:32], msgExt))


def separateParameters(paramSetList):
    """Separates paramSetList into parameters with single and multiple values"""
    singleVaules = {}
    multiValues = {}
    for name, values in paramSetList.items():
        if values == None:
            printExit("You must specify value(s) for parameter \"{}\"".format(name))
        if len(values) == 1 and name != "ProblemSizes":
            singleVaules[name] = values[0]
        elif len(values) > 1 and name != "ProblemSizes":
            multiValues[name] = values

    return singleVaules, multiValues


def checkCDBufferAndStrides(problemType, problemSizes, isCEqualD):
    """Ensures ldd == ldc when CEqualD"""
    if isCEqualD and problemType["OperationType"] == "GEMM":
        for problem in problemSizes.problems:
            ldd = problem.sizes[problemType["IndexAssignmentsLD"][0]]
            ldc = problem.sizes[problemType["IndexAssignmentsLD"][1]]
            if ldd != ldc:
                printExit("LDD({}) != LDC({}) causes unpredictable result when CEqualD(True)" \
                        .format(ldd, ldc))


class BenchmarkProcess:
    """Representation of benchmarking parameters and resulting steps"""

    def __init__(self, problemTypeConfig, problemSizeGroupConfig):
        """Create from the two sections of a config for a BenchmarkProblem"""
        self.problemType = ProblemType(problemTypeConfig)
        self.isBatched = "Batched" in problemTypeConfig and problemTypeConfig["Batched"]
        tPrint(3, "# BenchmarkProcess beginning {}".format(self.problemType))

        # fill parameter values from config
        self.singleValueParams = {}
        self.multiValueParams = {}
        self.customKernels = []
        self.sizes = None
        self.getConfigParameters(self.isBatched, problemSizeGroupConfig)

        # convert parameter lists to steps
        # previously, multiple benchmark steps were possible
        # currently only 1 benchmark step is possible; more may be added back later
        self.benchmarkSteps = []
        self.benchmarkStepIdx = 0
        self.convertParametersToSteps()

    def getConfigParameters(self, isbatched, config):
        """Parse and validate benchmarking parameters in config"""
        tPrint(3, "")
        tPrint(3, "####################################################################")
        tPrint(1, "# Filling in Parameters With Defaults")
        tPrint(3, "####################################################################")
        tPrint(3, "")

        # check for no longer supported legacy benchmark steps
        badParams = ["InitialSolutionParameters", "BenchmarkForkParameters", \
                     "JoinParameters", "BenchmarkJoinParameters"]
        badsInConfig = []

        for p in badParams:
            if config.get(p) is not None:
                badsInConfig.append(p)

        if len(badsInConfig) == 1:
            printExit("Benchmark step {} is no longer supported".format("'" + badsInConfig[0] +
                                                                        "'"))
        elif len(badsInConfig) > 1:
            printExit("Benchmark steps {} are no longer supported".format(badsInConfig))

        # get supported configurations
        # value in config file may be "None", which we should ignore
        def getNonNoneFromConfig(key, default):
            if config.get(key) is not None:
                return config[key]
            else:
                return default

        # converts list of dicts into a flat dict
        benchmarkCommonParams = dict(itertools.chain(*[x.items() \
                for x in getNonNoneFromConfig("BenchmarkCommonParameters", [])]))
        forkParams = dict(itertools.chain(*[x.items() \
                for x in getNonNoneFromConfig("ForkParameters", [])]))
        self.paramGroups = forkParams.pop("Groups") if "Groups" in forkParams else []
        self.customKernels = getNonNoneFromConfig("CustomKernels", [])

        if "BenchmarkFinalParameters" in config:
            sizes = config["BenchmarkFinalParameters"][0]["ProblemSizes"]
        else:
            sizes = defaultBatchedBenchmarkFinalProblemSizes if isbatched \
                else defaultBenchmarkFinalProblemSizes

        self.problemSizes = ProblemSizes(self.problemType, sizes)
        checkCDBufferAndStrides(self.problemType, self.problemSizes, globalParameters["CEqualD"])

        # validate parameter values
        configParams = {**benchmarkCommonParams, **forkParams}
        for param in configParams.items():
            checkParametersAreValid(param, validParameters)

        # TODO other checks on groups (same params for each entry? no dups between groups?)
        for list in self.paramGroups:
            for group in list:
                for k, v in group.items():
                    checkParametersAreValid((k, [v]), validParameters)

        params = dict(itertools.chain(*[x.items() for x in defaultBenchmarkCommonParameters]))
        params.update(configParams)
        self.singleValueParams, self.multiValueParams = separateParameters(params)

        # print summary of parameter values
        tPrint(3, "Single Value Parameters:")
        for k, v in sorted(self.singleValueParams.items()):
            tPrint(3, "    {}: {}".format(k, v))

        tPrint(3, "Multi-Value Parameters:")
        for k, v in sorted(self.multiValueParams.items()):
            tPrint(3, "    {}: {}".format(k, v))

        if len(self.paramGroups) > 0:
            tPrint(3, "{} Parameter Group(s):".format(len(self.paramGroups)))
            for i, group in enumerate(self.paramGroups):
                tPrint(3, "    {} entries is group {}".format(len(group), i + 1))

    def convertParametersToSteps(self):
        """Create benchmark steps based on parsed parameters"""
        tPrint(3, "")
        tPrint(3, "####################################################################")
        tPrint(1, "# Convert Parameters to Benchmark Step(s)")
        tPrint(3, "####################################################################")
        tPrint(3, "")

        # currently only a single step is supported
        tPrint(3, "")
        tPrint(3, "####################################################################")
        tPrint(1, "# Benchmark Final")
        benchmarkStep = BenchmarkStep( \
                self.multiValueParams, \
                self.singleValueParams, \
                self.paramGroups, \
                self.customKernels, \
                self.problemSizes, \
                self.benchmarkStepIdx)
        self.benchmarkSteps.append(benchmarkStep)
        self.benchmarkStepIdx += 1

    def __len__(self):
        return len(self.benchmarkSteps)

    def __getitem__(self, key):
        return self.benchmarkSteps[key]

    def __str__(self):
        string = "BenchmarkProcess:\n"
        for step in self.benchmarkSteps:
            string += str(step)
        return string

    def __repr__(self):
        return self.__str__()


def constructForkPermutations(forkParams, paramGroups):
    """Constructs cartesian product of parameter values in forkParams and paramGroups"""

    myParams = {}
    myParams.update(forkParams)

    totalPermutations = 1
    for _, values in forkParams.items():
        totalPermutations *= len(values)

    # add groups to parameters to fork on
    for i, group in enumerate(paramGroups):
        myParams["_group" + str(i)] = group
        totalPermutations *= len(group)

    forkPermutations = []
    for i in range(0, totalPermutations):
        permutation = {}
        pIdx = i
        for name, v in myParams.items():
            values = deepcopy(v)
            valueIdx = pIdx % len(v)

            # groups have multiple parameters to update
            if "_group" in name:
                entry = values[valueIdx]
                for n2, v2 in entry.items():
                    permutation[n2] = v2
            else:
                permutation[name] = values[valueIdx]

            pIdx //= len(values)
        forkPermutations.append(permutation)

    return forkPermutations


class BenchmarkStep:
    """A single benchmark step which consists of constant and fork parameters and a set of sizes"""

    def __init__(self, forkParams, constantParams, paramGroups, customKernels, problemSizes, idx):
        """Basic constructor storing each argument"""
        self.forkParams = forkParams
        self.constantParams = constantParams
        self.paramGroups = paramGroups
        self.customKernels = customKernels
        self.problemSizes = problemSizes
        self.stepIdx = idx

        self.customKernelWildcard = False
        if self.customKernels == ["*"]:
            self.customKernels = getAllCustomKernelNames()
            self.customKernelWildcard = True

        tPrint(3, "# Creating BenchmarkStep: {} fork params and {} sizes" \
                .format( len(forkParams), problemSizes.totalProblemSizes))

    def isFinal(self):
        """Legacy. Currently always returns true since only one benchmark step is possible"""
        return True

    def __str__(self):
        string = "{:02d}".format(self.stepIdx)
        if self.isFinal():
            string += "_Final"
        return string

    def __repr__(self):
        return self.__str__()



################################################################################
#
# Copyright (C) 2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from copy import deepcopy

from .Common import globalParameters, CHeader
from .KernelWriterBase import KernelWriterBase

class KernelWriterStreamKInit(KernelWriterBase):

  def __init__(self, state):
    super().__init__()

    self.state["ProblemType"] = deepcopy(state["ProblemType"])
    self.state["_GlobalAccumulation"] = state["_GlobalAccumulation"]

    # derive parameter
    self.language = "HIP"
    self.kernelName = self.getKernelName()


  def functionSignature(self):
    kStr = ""

    # self.state name
    kStr += self.endLine
    kStr += "extern \"C\"" + self.endLine
    kStr += "__global__ "
    kStr += "void %s" % ( self.kernelName )
    kStr += "(" + self.endLine

    # pointers
    kStr += " unsigned int * Flags," + self.endLine # Already offset to start of flags section in workspace

    kStr += " unsigned int const flagCount" + self.endLine

    kStr += " )%s" % (self.endLine)

    return kStr


  ##############################################################################
  # Kernel Body Stream-K Init
  ##############################################################################
  def kernelBodyStreamKInit(self):
    kStr = ""
    kStr += "{%s" % self.endLine

    ########################################
    # Stream-K initialize flags to 0
    kStr += "  uint64_t id = %s(0);%s" % (self.getGlobalIdStr, self.endLine)
    kStr += "  if (id >= (flagCount))" + self.endLine
    kStr += "    return;%s" % self.endLine
    kStr += self.endLine

    kStr += "  Flags[id] = 0;" + self.endLine

    ########################################
    # end
    kStr += "}%s" % self.endLine

    return kStr


  def getKernelName(self):
    # Output to workspace flags
    name = "WSFlags"
    # name += "_"
    # name += self.state["ProblemType"]["DestDataType"].toChar()
    return name


  def getSourceFileString(self):
    fileString = ""

    if not globalParameters["MergeFiles"]:
      fileString += "\n"
      fileString += "#include \"%s.h\"\n" % self.kernelName
      fileString += "\n"

    fileString += self.functionSignature()
    fileString += self.kernelBodyStreamKInit()

    return (0, fileString)

  def getHeaderFileString(self):
    fileString = "" # CHeader
    if not globalParameters["MergeFiles"]:
      fileString += CHeader
      fileString += "#pragma once\n\n"
      fileString += "\n"
      fileString += "#include <KernelHeader.h>\n\n"
      fileString += "#include <hip/hip_runtime.h>\n"
      fileString += "#include <hip/hip_fp16.h>\n"
      fileString += "\n"

    fileString += self.functionSignature()
    fileString += ";\n"

    return fileString



################################################################################
#
# Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import argparse
import csv
import pathlib
import sys

from collections import defaultdict

def gather_data(indexFile):

    indexData = defaultdict(set)
    with open(indexFile, "rt") as csvfile:
        indexreader = csv.DictReader(csvfile, delimiter=",")
        for row in indexreader: # read a row as {column1: value1, column2: value2,...}
            for key, value in row.items():
                indexData[key].add(value)
    return indexData

if __name__ == "__main__":
    argParser = argparse.ArgumentParser()
    argParser.add_argument("library_path", help="Tensile library path")
    args = argParser.parse_args()
    libraryPath = args.library_path
    indexFileName = "TensileMasterSolutionIndex.csv"

    # Check that path exists
    if not pathlib.Path(libraryPath).is_dir():
        print(f"ERROR: {libraryPath} does not exists.")
        sys.exit(1)

    # Check that TensileMasterSolutionIndex.csv exists
    csvpath = pathlib.Path(libraryPath) / indexFileName
    if not csvpath.is_file():
        print(f"ERROR: {csvpath} does not exists.")
        sys.exit(1)

    data = gather_data(csvpath)

    # List files in library path
    datFiles = [f.stem for f in pathlib.Path(libraryPath).glob("*.dat")]
    coFiles = [f.stem for f in pathlib.Path(libraryPath).glob("*.co")]
    lazyArchFiles = [f for f in datFiles if "_lazy_" in f]
    metaDataFiles = [f for f in datFiles if not "_lazy_" in f]
    nonfallback = set([f for f in data['libraryName'] if not "fallback" in f])

    print(f"MetaData files should match library names in index file: {set(metaDataFiles) == data['libraryName']}")
    print(f"Asm files should match non-fallback library names in index file: {set(coFiles) == nonfallback}")
    print(f"Lazy library files should match number of architectures in index file: {len(set(lazyArchFiles)) == len(data['architectureName'])}")



################################################################################
#
# Copyright (C) 2021-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import pytest
import os
from Tensile.CustomKernels import getCustomKernelConfig, getCustomKernelContents
from Tensile.BenchmarkProblems import getCustomKernelSolutionObj
from Tensile.Common import assignGlobalParameters
from Tensile.Utilities.ConditionalImports import yamlLoader
from Tensile.Utilities.Toolchain import validateToolchain, ToolchainDefaults

import yaml

testKernelDir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "customKernels")

@pytest.mark.parametrize("objs", [("TestKernel", testKernelDir)])
def test_FindCustomKernel(objs):
    try:
        name, directory = objs
        contents = getCustomKernelContents(name, directory)
        assert contents #If no exception
    except:
        assert False

configResult = yaml.load(
"""
ProblemType:
    OperationType: GEMM
    DataType: s
    TransposeA: False
    TransposeB: False
    UseBeta: True
    Batched: True
LoopDoWhile: False
WorkGroupMapping:  1
ThreadTile: [ 8, 8 ]
WorkGroup: [  8, 16,  1 ]
DepthU: 8
VectorWidth: 4
AssertSizeEqual: {3: 512}
AssertSizeMultiple: {0: 128, 1: 128}""", yamlLoader
)

# TODO when more custom kernels have been added - expand these lists
@pytest.mark.parametrize("objs", [("TestKernel", testKernelDir, configResult)])
def test_ReadCustomKernelConfig(objs):
    try:
        name, directory, result = objs
        config = getCustomKernelConfig(name, directory)
        config["custom.config"] = result
    except:
        assert False

@pytest.mark.parametrize("objs", [("TestKernel", testKernelDir)])
def test_CreateSolutionFromCustomKernel(objs):
    # try:
    (
        cxxCompiler,
        cCompiler,
        assembler,
        offloadBundler,
        hipconfig,
        # deviceEnumerator
    ) = validateToolchain(
        ToolchainDefaults.CXX_COMPILER,
        ToolchainDefaults.C_COMPILER,
        ToolchainDefaults.ASSEMBLER,
        ToolchainDefaults.OFFLOAD_BUNDLER,
        ToolchainDefaults.HIP_CONFIG,
        # ToolchainDefaults.DEVICE_ENUMERATOR
    )
    params = {
        "CxxCompiler": cxxCompiler,
        "CCompiler": cCompiler,
        "Assembler": assembler,
        "OffloadBundler": offloadBundler,
        "HipConfig": hipconfig,
        # "ROCmAgentEnumeratorPath": deviceEnumerator,
    }
    params["ROCmAgentEnumeratorPath"] = None
    assignGlobalParameters(params)

    name, directory = objs
    solution = getCustomKernelSolutionObj(name, directory)
    assert solution["Valid"]
    # except:
    #     assert False



################################################################################
#
# Copyright (C) 2021-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import ast
from Tensile.Configuration import ReadWriteTransformDict
from Tensile.Configuration import Parameter
from Tensile.Configuration import CallableParameter
from Tensile.Configuration import ExpressionEvaluator
from Tensile.Configuration import ProjectConfig

def test_ReadWriteTransformDict():
    def readXForm(obj, key):
        return obj.readNoTransform(key+"-Mod")

    def writeXForm(obj, key, value):
        obj.writeNoTransform(key+"-Mod", value)

    test = ReadWriteTransformDict(readXForm, writeXForm)

    # Test with the re-direction of transform funcs
    test["Key"] = 10
    assert test["Key"] == 10 and not "Key" in test
    assert test["Key"] == test.readWithTransform("Key")
    assert "Key-Mod" in test and test.readNoTransform("Key-Mod") == test["Key"]

    # Test bypass
    test.writeNoTransform("Key", 25)
    assert "Key" in test and test.readNoTransform("Key") == 25
    assert test["Key"] != test.readNoTransform("Key")

    # Test removal
    test.setReadTransform(None)
    test.setWriteTransform(None)

    # Test restoration of default behaviour
    test["Key"] = -5
    assert test["Key"] == -5 and test.readNoTransform("Key") == -5

def test_Parameter():
    name = "banana"
    initVal = 15
    defaultVal = 10
    descr = "I am a banana"
    initType = type(initVal)
    b = Parameter(name, initVal, defaultVal, descr)

    #attr
    assert b.value == initVal
    assert b.defaultValue == defaultVal
    assert b.name == name
    assert b.description == descr
    assert b.type == initType

    b.createAttr("color", "yellow")
    assert b.color == "yellow"

    # Accessor funcs
    assert b.getDefault() == defaultVal
    assert b.getValue() == initVal
    assert b.getDescription() == descr
    assert b.resetToDefault() == defaultVal

    # Write Xform
    try:
        # Unknown attr should not be writable
        b.peel
        raise Exception("Should not get here")
    except KeyError:
        pass
    except :
        assert 0, "Failed mutability test"

    try:
        # This should not be writable
        b.type = int
        raise Exception("Should not get here")
    except AttributeError:
        pass
    except :
        assert 0, "Failed type mutability test"

    try:
        # Writing different type should throw
        # b.value initialized with int
        b.value = "Dude"
        raise Exception("Should not get here")
    except AttributeError:
        pass
    except :
        assert 0, "Failed type assignment test"

    try:
        # Writing same type should succeed
        b.value = 55
    except :
        assert 0, "Failed assignment test"

    b.resetToDefault()

    # Comparators const, rhs
    assert (5 < b) == (5 < defaultVal)
    assert (5 <= b) == (5 <= defaultVal)
    assert (5 == b) == (5 == defaultVal)
    assert (5 != b) == (5 != defaultVal)
    assert (5 > b) == (5 > defaultVal)
    assert (5 >= b) == (5 >= defaultVal)

    # Comparators const, lhs
    assert (b < 5) == (defaultVal < 5)
    assert (b <= 5) == (defaultVal <= 5)
    assert (b == 5) == (defaultVal == 5)
    assert (b != 5) == (defaultVal != 5)
    assert (b > 5) == (defaultVal > 5)
    assert (b >= 5) == (defaultVal >= 5)

    defaultVal2 = 1
    a = Parameter("Apple", defaultVal2, defaultVal2, "I am an apple")
    # Comparators param, lhs
    assert (a < b) == (defaultVal2 < defaultVal)
    assert (a <= b) == (defaultVal2 <= defaultVal)
    assert (a == b) == (defaultVal2 == defaultVal)
    assert (a != b) == (defaultVal2 != defaultVal)
    assert (a > b) == (defaultVal2 > defaultVal)
    assert (a >= b) == (defaultVal2 >= defaultVal)

    assert (b < a) == (defaultVal < defaultVal2)
    assert (b <= a) == (defaultVal <= defaultVal2)
    assert (b == a) == (defaultVal == defaultVal2)
    assert (b != a) == (defaultVal != defaultVal2)
    assert (b > a) == (defaultVal > defaultVal2)
    assert (b >= a) == (defaultVal >= defaultVal2)

    # Bin op const, rhs
    assert (5 + b) == (5 + defaultVal)
    assert (5 - b) == (5 - defaultVal)
    assert (5 * b) == (5 * defaultVal)
    assert (5 / b) == (5 / defaultVal)
    assert (5 // b) == (5 // defaultVal)
    assert (5 % b) == (5 % defaultVal)
    assert (5 ** b) == (5 ** defaultVal)
    assert (5 >> b) == (5 >> defaultVal)
    assert (5 << b) == (5 << defaultVal)
    assert (5 & b) == (5 & defaultVal)
    assert (5 | b) == (5 | defaultVal)
    assert (5 ^ b) == (5 ^ defaultVal)

    # Bin op const, lhs
    assert (b + 5) == (defaultVal + 5)
    assert (b - 5) == (defaultVal - 5)
    assert (b * 5) == (defaultVal * 5)
    assert (b / 5) == (defaultVal / 5)
    assert (b // 5) == (defaultVal // 5)
    assert (b % 5) == (defaultVal % 5)
    assert (b ** 5) == (defaultVal ** 5)
    assert (b >> 5) == (defaultVal >> 5)
    assert (b << 5) == (defaultVal << 5)
    assert (b & 5) == (defaultVal & 5)
    assert (b | 5) == (defaultVal | 5)
    assert (b ^ 5) == (defaultVal ^ 5)

    # Bin op param, rhs
    assert (a + b) == (defaultVal2 + defaultVal)
    assert (a - b) == (defaultVal2 - defaultVal)
    assert (a * b) == (defaultVal2 * defaultVal)
    assert (a / b) == (defaultVal2 / defaultVal)
    assert (a // b) == (defaultVal2 // defaultVal)
    assert (a % b) == (defaultVal2 % defaultVal)
    assert (a ** b) == (defaultVal2 ** defaultVal)
    assert (a >> b) == (defaultVal2 >> defaultVal)
    assert (a << b) == (defaultVal2 << defaultVal)
    assert (a & b) == (defaultVal2 & defaultVal)
    assert (a | b) == (defaultVal2 | defaultVal)
    assert (a ^ b) == (defaultVal2 ^ defaultVal)

    # Bin op param, lhs
    assert (b + a) == (defaultVal + defaultVal2)
    assert (b - a) == (defaultVal - defaultVal2)
    assert (b * a) == (defaultVal * defaultVal2)
    assert (b / a) == (defaultVal / defaultVal2)
    assert (b // a) == (defaultVal // defaultVal2)
    assert (b % a) == (defaultVal % defaultVal2)
    assert (b ** a) == (defaultVal ** defaultVal2)
    assert (b >> a) == (defaultVal >> defaultVal2)
    assert (b << a) == (defaultVal << defaultVal2)
    assert (b & a) == (defaultVal & defaultVal2)
    assert (b | a) == (defaultVal | defaultVal2)
    assert (b ^ a) == (defaultVal ^ defaultVal2)

    # Unary
    assert (~b) == (~defaultVal)
    assert (not b) == (not defaultVal)
    assert (-b) == (-defaultVal)
    assert (+b) == (+defaultVal)
    assert bool(b) == bool(defaultVal)

def test_CallableParameter():
    # Test with function
    def bananaFunc(obj):
        return obj.testVal
    b = CallableParameter("banana", bananaFunc, "I am a banana")
    b.createAttr("testVal", -99)

    assert b.__call__() == -99
    assert b.value == -99

    # Test with lambda
    appleFunc = lambda obj: obj.testVal
    a = CallableParameter("apple", appleFunc, "I am an apple")
    a.createAttr("testVal", 101)

    assert a.__call__() == 101
    assert a.value == 101

    # Test write
    a.value = 55
    assert a.value == 101

def test_BinaryOp():
    # Test 2 const
    lhs = -3
    rhs = 5

    p0 = CallableParameter.createBinaryOp(lhs, rhs, "And")
    p1 = CallableParameter.createBinaryOp(lhs, rhs, "Or")
    p2 = CallableParameter.createBinaryOp(lhs, rhs, "Lt")
    p3 = CallableParameter.createBinaryOp(lhs, rhs, "LtE")
    p4 = CallableParameter.createBinaryOp(lhs, rhs, "Eq")
    p5 = CallableParameter.createBinaryOp(lhs, rhs, "NotEq")
    p6 = CallableParameter.createBinaryOp(lhs, rhs, "Gt")
    p7 = CallableParameter.createBinaryOp(lhs, rhs, "GtE")
    p8 = CallableParameter.createBinaryOp(lhs, rhs, "Mult")
    p9 = CallableParameter.createBinaryOp(lhs, rhs, "Pow")
    p10 = CallableParameter.createBinaryOp(lhs, rhs, "Div")
    p11 = CallableParameter.createBinaryOp(lhs, rhs, "FloorDiv")
    p12 = CallableParameter.createBinaryOp(lhs, rhs, "Mod")
    p13 = CallableParameter.createBinaryOp(lhs, rhs, "Add")
    p14 = CallableParameter.createBinaryOp(lhs, rhs, "Sub")
    p15 = CallableParameter.createBinaryOp(lhs, rhs, "BitAnd")
    p16 = CallableParameter.createBinaryOp(lhs, rhs, "BitOr")
    p17 = CallableParameter.createBinaryOp(lhs, rhs, "min")
    p18 = CallableParameter.createBinaryOp(lhs, rhs, "max")

    def customOp0(lhs, rhs):
        return rhs

    customOp1 = lambda lhs, rhs : rhs

    def badOp(lhs):
        return lhs

    p19 = CallableParameter.createBinaryOp(lhs, rhs, customOp0)
    assert p19.name == "CustomBinaryOp"

    p20 = CallableParameter.createBinaryOp(lhs, rhs, customOp1)
    assert p20.name == "CustomBinaryOp"

    try:
        p21 = CallableParameter.createBinaryOp(lhs, rhs, badOp)
        p21.value # Dummy to satisfy linter
    except CallableParameter.BadFunc:
        pass
    except :
        assert 0, "Failed bad binary op function handling"

    assert p0.value == (lhs and rhs)
    assert p1.value == (lhs or rhs)
    assert p2.value == (lhs < rhs)
    assert p3.value == (lhs <= rhs)
    assert p4.value == (lhs == rhs)
    assert p5.value == (lhs != rhs)
    assert p6.value == (lhs > rhs)
    assert p7.value == (lhs >= rhs)
    assert p8.value == (lhs * rhs)
    assert p9.value == (lhs ** rhs)
    assert p10.value == (lhs / rhs)
    assert p11.value == (lhs // rhs)
    assert p12.value == (lhs % rhs)
    assert p13.value == (lhs + rhs)
    assert p14.value == (lhs - rhs)
    assert p15.value == (lhs & rhs)
    assert p16.value == (lhs | rhs)
    assert p17.value == min(lhs, rhs)
    assert p18.value == max(lhs, rhs)
    assert p19.value == rhs
    assert p20.value == rhs

    # Test lhs const
    lhs = -3
    rhs = Parameter("banana", 5, 5, "I am a banana")

    p0 = CallableParameter.createBinaryOp(lhs, rhs, "And")
    p1 = CallableParameter.createBinaryOp(lhs, rhs, "Or")
    p2 = CallableParameter.createBinaryOp(lhs, rhs, "Lt")
    p3 = CallableParameter.createBinaryOp(lhs, rhs, "LtE")
    p4 = CallableParameter.createBinaryOp(lhs, rhs, "Eq")
    p5 = CallableParameter.createBinaryOp(lhs, rhs, "NotEq")
    p6 = CallableParameter.createBinaryOp(lhs, rhs, "Gt")
    p7 = CallableParameter.createBinaryOp(lhs, rhs, "GtE")
    p8 = CallableParameter.createBinaryOp(lhs, rhs, "Mult")
    p9 = CallableParameter.createBinaryOp(lhs, rhs, "Pow")
    p10 = CallableParameter.createBinaryOp(lhs, rhs, "Div")
    p11 = CallableParameter.createBinaryOp(lhs, rhs, "FloorDiv")
    p12 = CallableParameter.createBinaryOp(lhs, rhs, "Mod")
    p13 = CallableParameter.createBinaryOp(lhs, rhs, "Add")
    p14 = CallableParameter.createBinaryOp(lhs, rhs, "Sub")
    p15 = CallableParameter.createBinaryOp(lhs, rhs, "BitAnd")
    p16 = CallableParameter.createBinaryOp(lhs, rhs, "BitOr")
    p17 = CallableParameter.createBinaryOp(lhs, rhs, "min")
    p18 = CallableParameter.createBinaryOp(lhs, rhs, "max")

    def customOp2(lhs, rhs):
        return lhs

    customOp3 = lambda lhs, rhs : lhs

    p19 = CallableParameter.createBinaryOp(lhs, rhs, customOp2)
    p20 = CallableParameter.createBinaryOp(lhs, rhs, customOp3)

    assert p0.value == (lhs and rhs)
    assert p1.value == (lhs or rhs)
    assert p2.value == (lhs < rhs)
    assert p3.value == (lhs <= rhs)
    assert p4.value == (lhs == rhs)
    assert p5.value == (lhs != rhs)
    assert p6.value == (lhs > rhs)
    assert p7.value == (lhs >= rhs)
    assert p8.value == (lhs * rhs)
    assert p9.value == (lhs ** rhs)
    assert p10.value == (lhs / rhs)
    assert p11.value == (lhs // rhs)
    assert p12.value == (lhs % rhs)
    assert p13.value == (lhs + rhs)
    assert p14.value == (lhs - rhs)
    assert p15.value == (lhs & rhs)
    assert p16.value == (lhs | rhs)
    assert p17.value == min(lhs, rhs)
    assert p18.value == max(lhs, rhs)
    assert p19.value == lhs
    assert p20.value == lhs

    # Test rhs const
    lhs = Parameter("banana", -3, -3, "I am a banana")
    rhs = 5

    p0 = CallableParameter.createBinaryOp(lhs, rhs, "And")
    p1 = CallableParameter.createBinaryOp(lhs, rhs, "Or")
    p2 = CallableParameter.createBinaryOp(lhs, rhs, "Lt")
    p3 = CallableParameter.createBinaryOp(lhs, rhs, "LtE")
    p4 = CallableParameter.createBinaryOp(lhs, rhs, "Eq")
    p5 = CallableParameter.createBinaryOp(lhs, rhs, "NotEq")
    p6 = CallableParameter.createBinaryOp(lhs, rhs, "Gt")
    p7 = CallableParameter.createBinaryOp(lhs, rhs, "GtE")
    p8 = CallableParameter.createBinaryOp(lhs, rhs, "Mult")
    p9 = CallableParameter.createBinaryOp(lhs, rhs, "Pow")
    p10 = CallableParameter.createBinaryOp(lhs, rhs, "Div")
    p11 = CallableParameter.createBinaryOp(lhs, rhs, "FloorDiv")
    p12 = CallableParameter.createBinaryOp(lhs, rhs, "Mod")
    p13 = CallableParameter.createBinaryOp(lhs, rhs, "Add")
    p14 = CallableParameter.createBinaryOp(lhs, rhs, "Sub")
    p15 = CallableParameter.createBinaryOp(lhs, rhs, "BitAnd")
    p16 = CallableParameter.createBinaryOp(lhs, rhs, "BitOr")
    p17 = CallableParameter.createBinaryOp(lhs, rhs, "min")
    p18 = CallableParameter.createBinaryOp(lhs, rhs, "max")

    def customOp4(lhs, rhs):
        return lhs+rhs*lhs

    customOp5 = lambda lhs, rhs : lhs+rhs*lhs

    p19 = CallableParameter.createBinaryOp(lhs, rhs, customOp4)
    p20 = CallableParameter.createBinaryOp(lhs, rhs, customOp5)

    assert p0.value == (lhs and rhs)
    assert p1.value == (lhs or rhs)
    assert p2.value == (lhs < rhs)
    assert p3.value == (lhs <= rhs)
    assert p4.value == (lhs == rhs)
    assert p5.value == (lhs != rhs)
    assert p6.value == (lhs > rhs)
    assert p7.value == (lhs >= rhs)
    assert p8.value == (lhs * rhs)
    assert p9.value == (lhs ** rhs)
    assert p10.value == (lhs / rhs)
    assert p11.value == (lhs // rhs)
    assert p12.value == (lhs % rhs)
    assert p13.value == (lhs + rhs)
    assert p14.value == (lhs - rhs)
    assert p15.value == (lhs & rhs)
    assert p16.value == (lhs | rhs)
    assert p17.value == min(lhs, rhs)
    assert p18.value == max(lhs, rhs)
    assert p19.value == (lhs+lhs*rhs)
    assert p20.value == (lhs+lhs*rhs)

    # Test lhs, rhs params
    lhs = Parameter("banana", -3, -3, "I am a banana")
    rhs = Parameter("apple", 5, 5, "I am an apple")

    p0 = CallableParameter.createBinaryOp(lhs, rhs, "And")
    p1 = CallableParameter.createBinaryOp(lhs, rhs, "Or")
    p2 = CallableParameter.createBinaryOp(lhs, rhs, "Lt")
    p3 = CallableParameter.createBinaryOp(lhs, rhs, "LtE")
    p4 = CallableParameter.createBinaryOp(lhs, rhs, "Eq")
    p5 = CallableParameter.createBinaryOp(lhs, rhs, "NotEq")
    p6 = CallableParameter.createBinaryOp(lhs, rhs, "Gt")
    p7 = CallableParameter.createBinaryOp(lhs, rhs, "GtE")
    p8 = CallableParameter.createBinaryOp(lhs, rhs, "Mult")
    p9 = CallableParameter.createBinaryOp(lhs, rhs, "Pow")
    p10 = CallableParameter.createBinaryOp(lhs, rhs, "Div")
    p11 = CallableParameter.createBinaryOp(lhs, rhs, "FloorDiv")
    p12 = CallableParameter.createBinaryOp(lhs, rhs, "Mod")
    p13 = CallableParameter.createBinaryOp(lhs, rhs, "Add")
    p14 = CallableParameter.createBinaryOp(lhs, rhs, "Sub")
    p15 = CallableParameter.createBinaryOp(lhs, rhs, "BitAnd")
    p16 = CallableParameter.createBinaryOp(lhs, rhs, "BitOr")
    p17 = CallableParameter.createBinaryOp(lhs, rhs, "min")
    p18 = CallableParameter.createBinaryOp(lhs, rhs, "max")

    def customOp6(lhs, rhs):
        return lhs-rhs*lhs

    customOp7 = lambda lhs, rhs : lhs-rhs*lhs

    p19 = CallableParameter.createBinaryOp(lhs, rhs, customOp6)
    p20 = CallableParameter.createBinaryOp(lhs, rhs, customOp7)

    assert p0.value == (lhs and rhs)
    assert p1.value == (lhs or rhs)
    assert p2.value == (lhs < rhs)
    assert p3.value == (lhs <= rhs)
    assert p4.value == (lhs == rhs)
    assert p5.value == (lhs != rhs)
    assert p6.value == (lhs > rhs)
    assert p7.value == (lhs >= rhs)
    assert p8.value == (lhs * rhs)
    assert p9.value == (lhs ** rhs)
    assert p10.value == (lhs / rhs)
    assert p11.value == (lhs // rhs)
    assert p12.value == (lhs % rhs)
    assert p13.value == (lhs + rhs)
    assert p14.value == (lhs - rhs)
    assert p15.value == (lhs & rhs)
    assert p16.value == (lhs | rhs)
    assert p17.value == min(lhs, rhs)
    assert p18.value == max(lhs, rhs)
    assert p19.value == (lhs-lhs*rhs)
    assert p20.value == (lhs-lhs*rhs)

def test_UnaryOp():
    # Test const
    rhs = 5

    p0 = CallableParameter.createUnaryOp(rhs, "Not")
    p1 = CallableParameter.createUnaryOp(rhs, "Invert")
    p2 = CallableParameter.createUnaryOp(rhs, "None")

    def customOp(rhs):
        return rhs*5

    customOp1 = lambda rhs : rhs*5

    def badOp(lhs, rhs):
        return lhs + rhs

    p3 = CallableParameter.createUnaryOp(rhs, customOp)
    assert p3.name == "CustomUnaryOp"

    p4 = CallableParameter.createUnaryOp(rhs, customOp1)
    assert p4.name == "CustomUnaryOp"

    try:
        p5 = CallableParameter.createUnaryOp(rhs, badOp)
        p5.value # Dummy to satisfy linter
    except CallableParameter.BadFunc:
        pass
    except :
        assert 0, "Failed bad binary op function"

    assert p0.value == (not rhs)
    assert p1.value == ~rhs
    assert p2.value == rhs
    assert p3.value == rhs*5
    assert p4.value == rhs*5

    # Test param
    rhs = Parameter("banana", 5, 5, "I am a banana")

    p0 = CallableParameter.createUnaryOp(rhs, "Not")
    p1 = CallableParameter.createUnaryOp(rhs, "Invert")
    p2 = CallableParameter.createUnaryOp(rhs, "None")

    def customOp(rhs):
        return rhs*5

    customOp1 = lambda rhs : rhs*5

    p3 = CallableParameter.createUnaryOp(rhs, customOp)
    assert p3.name == "CustomUnaryOp"

    p4 = CallableParameter.createUnaryOp(rhs, customOp1)
    assert p4.name == "CustomUnaryOp"

    assert p0.value == (not rhs)
    assert p1.value == ~rhs
    assert p2.value == rhs
    assert p3.value == rhs*5
    assert p4.value == rhs*5

def test_ExpressionEvaluator():
    # Context
    a = 5
    b = -5
    c = 1
    d = -2
    context = {"a": a, "b": b, "c": c, "d": d}

    # Binary ops
    expr = "a + (b + (c*d) % a - 25)//d + 20"

    tree = ast.parse(expr)
    exprEval = ExpressionEvaluator().evaluate(tree, context)
    assert  exprEval == (a + (b + (c*d) % a - 25)//d + 20)

    # Mixed ops
    expr = "+a < (-b ^ d + ((c << 1) * (d >> 5)) > ~20 >= (not a & c | d) and c <= d or a == d)"
    tree = ast.parse(expr)
    exprEval = ExpressionEvaluator().evaluate(tree, context)
    assert  exprEval == (+a < (-b ^ d + ((c << 1) * (d >> 5)) > ~20 >= (not a & c | d) and c <= d or a ==d))

    # Assignment to value
    expr = "a=20"
    tree = ast.parse(expr)
    exprEval = ExpressionEvaluator().evaluate(tree, context)
    assert(context["a"] == 20) and (exprEval == 20)

    # Assignment to name
    expr = "a=b"
    tree = ast.parse(expr)
    exprEval = ExpressionEvaluator().evaluate(tree, context)
    assert(context["a"] == context["b"]) and (exprEval == context["b"])

    # Multiple assignment
    expr = "a=b=c"
    tree = ast.parse(expr)
    exprEval = ExpressionEvaluator().evaluate(tree, context)
    assert(context["a"] == context["c"]) and (context["b"] == context["c"]) and (exprEval == context["c"])

    # Assignment to expression
    expr = "b = max(5, 10)*10 / (2**4) - min(b, 4 % 2//1) if (1 != 2) else 0"
    compareVal = max(5, 10)*10 / (2**4) - min(context["b"], 4 % 2//1) if (1 != 2) else 0
    tree = ast.parse(expr)
    exprEval = ExpressionEvaluator().evaluate(tree, context)
    assert (context["b"] == compareVal) and (exprEval == compareVal)


def test_ProjectConfig():
    proj = ProjectConfig()
    proj.createValue("tVal", 75, -99, "TestDescription")
    subA = proj.createSection("SubA")
    subA.createValue("tVal", 0)
    subA.createValue("tMin", 0)
    subA.createValue("tMax", 99)
    subB = proj.createSection("SubB")
    subB.createValue("tVal", 0)
    subB.createValue("tMin", 78)
    subB.createValue("tMax", 100)
    subC = subB.createSection("SubC")
    subC.createValue("tVal", 0, 500, "Test")
    subC.createValue("tMin", -10000)
    subC.createValue("tMax", -500)

    # Assertions with different hierarchy levels
    proj.addConstraint("tVal >= SubA.tMin and tVal >= SubB.SubC.tMax")

    # Hard assignments for constraints
    proj.addConstraint("SubA.tVal = max(SubA.tMin, min(SubA.tMax, tVal))")
    proj.addConstraint("SubB.tVal = max(SubB.tMin, min(SubB.tMax, tVal))")
    proj.addConstraint("SubB.SubC.tVal = max(SubB.SubC.tMin, min(SubB.SubC.tMax, tVal))")
    proj.addConstraint("tVal = max(SubA.tVal, max(SubB.tVal, SubB.SubC.tVal))")
    proj.checkConstraints()

    assert subA.tVal == 75
    assert subB.tVal == 78
    assert subC.tVal == -500
    assert proj.tVal == 78

    # Access patterns:
    # Attribute chains. A.B.C
    # Item chains ["A"]["B"]["C"]
    # Item attribute chains ["A.B.C"]
    proj.SubA.tVal = 50
    assert subA.tVal == 50
    proj["SubA.tVal"] = 29
    assert subA.tVal == 29

    proj.tVal = -1
    assert proj.tVal == -1
    proj["tVal"] = +2
    assert proj.tVal == +2

    proj.SubB.SubC.tVal = 44
    assert subC.tVal == 44
    proj["SubB.SubC.tVal"] = 21
    assert subC.tVal == 21
    proj["SubB"]["SubC"]["tVal"] = 42
    assert subC.tVal == 42

    assert proj.getDefaultValue("tVal") == -99
    assert proj.getDescription("tVal") == "TestDescription"

    assert proj.getDefaultValue("SubB.SubC.tVal") == 500
    assert proj.getDescription("SubB.SubC.tVal") == "Test"



################################################################################
#
# Copyright (C) 2021-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import Component, MAC
from ..DataType import DataType


class FMA_I8_HPA_DOT4(MAC):
    asmCaps = lambda caps: caps["v_dot4c_i32_i8"] or caps["v_dot4_i32_i8"]
    kernel = {"ProblemType": {"DataType": DataType(DataType.int8),
                              "HighPrecisionAccumulate": True},
              "LocalDotLayout": 4
             }

    def __call__(self, writer, m, innerUnroll):
        kernel = writer.kernel

        kStr = self.commentHeader()

        priority = Component.Priority.find(writer)

        endLine     = writer.endLine
        accumulate  = writer.asmCaps["v_dot4c_i32_i8"]
        ThreadTile0 = kernel["ThreadTile0"]
        ThreadTile1 = kernel["ThreadTile1"]
        instruction = "v_dot4c_i32_i8" if accumulate else "v_dot4_i32_i8"
        inTSize     = 4

        kStr += '// C index : {blockA}*{inTSize} + {blockB}*{inTSize}*{ThreadTile0} + {indexB}*{ThreadTile0} + {indexA}' + endLine

        for blockB in range(0, ThreadTile1//inTSize):
            for blockA in range(0, ThreadTile0//inTSize):
                for indexB in range(inTSize):
                    for indexA in range(inTSize):
                        cIdxStr = f'{blockA}*{inTSize} + {blockB}*{inTSize}*{ThreadTile0} + {indexB}*{ThreadTile0} + {indexA}'
                        cIdx = eval(cIdxStr)

                        cStr = f'v[vgprValuC + {cIdxStr}]'
                        aStr = f'v[vgprValuA_X{m}_I{indexA}+{blockA}]'
                        bStr = f'v[vgprValuB_X{m}_I{indexB}+{blockB}]'
                        if accumulate:
                          kStr += f'{instruction} {cStr}, {aStr}, {bStr} // ValuC[{cIdx}]{endLine}'
                        else:
                          kStr += f'{instruction} {cStr}, {aStr}, {bStr}, {cStr} // ValuC[{cIdx}]{endLine}'

                        kStr += priority(writer, 1, "Raise priority while processing macs")

        kStr += priority(writer, 0, "Reset priority after macs")

        return kStr


class FMA_I8_HPA(MAC):
    @staticmethod
    def asmCaps(caps):
        return True

    kernel = {
        "ProblemType": {"DataType": DataType(DataType.int8), "HighPrecisionAccumulate": True},
        "LocalDotLayout": 1
    }

    def __call__(self, writer, m, innerUnroll):
        kernel      = writer.kernel
        priority    = Component.Priority.find(writer)
        spacePerReg = writer.bpr // writer.bpeAB
        elemPerReg  = min(kernel['VectorWidth'], spacePerReg)
        endLine     = writer.endLine

        kStr = self.commentHeader()

        for a in range(kernel["ThreadTile0"]-1, -1, -1):
            for iui in range(0, innerUnroll):
                src  = a // elemPerReg
                idx  = a %  elemPerReg
                sStr = f'vgprValuA_X{m}_I{iui}+{src}'
                tStr = f'vgprValuA_X{m}_I{iui}+{a}'
                kStr += f'v_lshlrev_b32 v[{tStr}], {(spacePerReg-idx-1)*8}, v[{sStr}]{endLine}'
                kStr += priority(writer, 1, "Raise priority while processing macs")
                kStr += f'v_ashrrev_i32 v[{tStr}], {(spacePerReg    -1)*8}, v[{tStr}]{endLine}'

        for b in range(kernel["ThreadTile1"]-1, -1, -1):
            for iui in range(0, innerUnroll):
                src  = b // elemPerReg
                idx  = b %  elemPerReg
                sStr = f'vgprValuB_X{m}_I{iui}+{src}'
                tStr = f'vgprValuB_X{m}_I{iui}+{b}'
                kStr += f'v_lshlrev_b32 v[{tStr}], {(spacePerReg-idx-1)*8}, v[{sStr}]{endLine}'
                kStr += priority(writer, 1, "Raise priority while processing macs")
                kStr += f'v_ashrrev_i32 v[{tStr}], {(spacePerReg    -1)*8}, v[{tStr}]{endLine}'

        ThreadTile0 = kernel["ThreadTile0"]
        for b in range(0, kernel["ThreadTile1"]):
            for a in range(0, kernel["ThreadTile0"]):
                for iui in range(0, innerUnroll):
                    cStr = f'v[vgprValuC + {a} + {b}*{ThreadTile0}]'
                    aStr = f'v[vgprValuA_X{m}_I{iui} + {a}]'
                    bStr = f'v[vgprValuB_X{m}_I{iui} + {b}]'
                    kStr += f'v_mad_i32_i24 {cStr}, {aStr}, {bStr}, {cStr}{endLine}'
                    kStr += priority(writer, 1, "Raise priority while processing macs")

        kStr += priority(writer, 0, "Reset priority after macs")

        return kStr



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import logging,pytest
from Tensile.SolutionStructs import Convolution
log =logging.getLogger("testlog")

@pytest.mark.skip(reason="backward_data under construction")
def test_nchw_backwarddata_defaults(tensile_state, run_convolution_level):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionBackwardData',
              config={'TensorAFormat': 'NCHW',
                      'Spatial' : '14x14',
                      })
    log.debug(conv.printUsage(z))
    if not tensile_state.args["no_conv_assertions"]:
        assert(z['NumIndicesC']==2)
        assert(z['IndexAssignmentsA']==[3, 0, 2])
        assert(z['IndexAssignmentsB']==[3, 1, 2])
        #assert(conv.solutionParms["AssertStrideAEqual"] == "1:1,0:1")
        #assert(conv.solutionParms["AssertStrideBEqual"] == "1:1")
        assert(conv.solutionParms["AssertSizeEqual"] == {})
    run_convolution_level.func(conv, z, run_convolution_level.solution)

@pytest.mark.skip(reason="backward_data under construction")
def test_nchw_backwarddata_filter3x1(tensile_state, run_convolution_level):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionBackwardData',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '3x1',
                      })
    log.debug(conv.printUsage(z))
    if not tensile_state.args["no_conv_assertions"]:
        assert(z['NumIndicesC']==3)
        assert(z['IndexAssignmentsA']==[4, 0, 1, 3])
        assert(z['IndexAssignmentsB']==[4, 2, 3])
        #assert(conv.solutionParms["AssertStrideAEqual"] == "2:1,0:1")
        #assert(conv.solutionParms["AssertStrideBEqual"] == "1:1")
        assert(conv.solutionParms["AssertSizeEqual"] == {4:1})
    run_convolution_level.func(conv, z, run_convolution_level.solution)

@pytest.mark.skip(reason="backward_data under construction")
def test_nchw_backwarddata_filter1x3(tensile_state, run_convolution_level):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionBackwardData',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '1x3',
                      })
    log.debug(conv.printUsage(z))
    if not tensile_state.args["no_conv_assertions"]:
        assert(z['NumIndicesC']==3)
        assert(z['IndexAssignmentsA']==[4, 0, 1, 3])
        assert(z['IndexAssignmentsB']==[4, 2, 3])
        #assert(conv.solutionParms["AssertStrideAEqual"] == "1:1,2:1,0:1")
        #assert(conv.solutionParms["AssertStrideBEqual"] == "1:1")
        assert(conv.solutionParms["AssertSizeEqual"] == {4:1})
    run_convolution_level.func(conv, z, run_convolution_level.solution)

@pytest.mark.skip(reason="backward_data under construction")
def test_nchw_backwarddata_filter3x5(tensile_state, run_convolution_level):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionBackwardData',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '3x5',
                      })
    log.debug(conv.printUsage(z))
    if not tensile_state.args["no_conv_assertions"]:
        assert(z['NumIndicesC']==4)
        assert(z['IndexAssignmentsA']==[5, 0, 1, 2, 4])
        assert(z['IndexAssignmentsB']==[5, 3, 4])
        #assert(conv.solutionParms["AssertStrideAEqual"] == "1:1,3:1,0:1")
        #assert(conv.solutionParms["AssertStrideBEqual"] == "1:1")
        assert(conv.solutionParms["AssertSizeEqual"] == {4:3,5:5})
    run_convolution_level.func(conv, z, run_convolution_level.solution)

def test_nchw_backwarddata_filter3x5_nopack(tensile_state, run_convolution_level):
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionBackwardData',
              config={'TensorAFormat': 'NCHW',
                      'Filter': '3x5',
                      'PackedSpatialDims': 0,
                      })
    log.debug(conv.printUsage(z))
    if not tensile_state.args["no_conv_assertions"]:
        (cdim,filterDims) = (6,[5,4]) if conv.unrollOnChannel else (4,[6,5])
        assert(z['NumIndicesC']==4)
        assert(z['IndexAssignmentsA'] == filterDims + [0, 1, cdim, 3])
        assert(z['IndexAssignmentsB'] == filterDims + [2, cdim, 3])
        assert(conv.solutionParms["AssertStrideAEqual"] == {0:1, 2:1})
        #assert(conv.solutionParms["AssertStrideBEqual"] == {0:1, filterDims[1]:0})
        assert(conv.solutionParms["AssertSizeEqual"] == {filterDims[1]:3,filterDims[0]:5})
    run_convolution_level.func(conv, z, run_convolution_level.solution)



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import pytest
from Tensile.SolutionStructs import Convolution
from YamlBuilder.YamlBuilder import YamlBuilder
from YamlBuilder.YamlBuilder import defaultSizes, resnetSizes

@pytest.mark.parametrize(
        "problemSizes",
        [pytest.param((YamlBuilder.ProblemSizes,level), id="default-lvl=%d"%level) for level in [1,2,3,4]] +
        [resnetSizes]
        )
def test_2d_stride1(problemSizes):
    "Test number of problems generated"
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Stride': '1x1',
                      'Filter': '1x1',
                      })

    exacts = problemSizes[0](conv, z, problemSizes[1])
    if exacts:
        None


@pytest.mark.parametrize("problemSizes", [defaultSizes, resnetSizes])
@pytest.mark.parametrize("problemLevel", [1,2,3,4])
def test_2d_stride2(problemSizes, problemLevel):
    "Test number of problems generated"
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Stride': '2x2',
                      'Filter': '3x3',
                      })

    exacts = problemSizes[0](conv, z, problemLevel)
    if exacts:
        None


@pytest.mark.parametrize("problem_level", [1,2,3,4])
def test_3d(problem_level):
    "Test number of problems generated"
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCDHW',
                      'Stride': '2x2x2',
                      'Filter': '3x3x3',
                      })

    exacts = YamlBuilder.ProblemSizes(conv, z, problem_level)
    if exacts:
        None



################################################################################
#
# Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from Tensile.AsmRegisterPool import RegisterPool
from Tensile.Utils import roundUpToNearestMultiple

S = RegisterPool.Status


def test_RegisterPool_add0():
    poolSize = 12

    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=False, printRP=0)
    for reg in vgprPool.pool:
        assert reg.status == S.Unavailable
        assert reg.tag == "init"

    vgprPool.add(0, poolSize, "tag")
    assert len(vgprPool.pool) == poolSize
    for reg in vgprPool.pool:
        assert reg.status == S.Available
        assert reg.tag == "tag"


def test_RegisterPool_add1():
    poolSize = 12
    start = 8
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=False, printRP=0)

    vgprPool.add(start, poolSize, "tag")
    assert len(vgprPool.pool) == poolSize + start
    for i, reg in enumerate(vgprPool.pool):
        if i < start:
            assert reg.status == S.Unavailable
            assert reg.tag == "init"
        else:
            assert reg.status == S.Available
            assert reg.tag == "tag"


def test_findFreeRange_vgprEndToEnd1_AllAvailable():
    poolSize = 12
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=False, printRP=0)

    vgprPool.add(0, poolSize, "tag")
    out_new = vgprPool.findFreeRange(4, 8)
    out_old = findFreeRange_oldLogic(vgprPool, 4, 8)

    assert out_new == out_old == 0


def test_findFreeRange_noOverflowButAvailable():
    poolSize = 12
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=True, printRP=0)
    vgprPool.add(6, poolSize, "tag")

    # Starting at index 3, find 4 free registers and return the starting index
    out_new = vgprPool.findFreeRange(3, 4)
    out_old = findFreeRange_oldLogic(vgprPool, 3, 4)

    assert out_new == out_old == 8


def test_findFreeRange_variableOveflowNotAvailable():
    poolSize = 12
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=False, printRP=0)

    vgprPool.add(0, 3, "tag")
    vgprPool.add(10, 2, "tag")

    #   A       A       A
    # [ . . . | | | | | | | . . ]
    #   0                     11
    out_new = vgprPool.findFreeRange(4, 4, preventOverflow=True)
    out_old = findFreeRange_oldLogic(vgprPool, 4, 4, preventOverflow=True)

    # We expect to hit the preventOverflow here
    assert out_new == out_old == None

    out_new = vgprPool.findFreeRange(4, 4)
    out_old = findFreeRange_oldLogic(vgprPool, 4, 4)

    # We expect to get values from the call to `startOfLastAvailableBlock`
    assert out_new == out_old == 12


def test_findFreeRange_noOverflowNotAvailable():
    poolSize = 12
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=True, printRP=0)

    vgprPool.add(7, 5, "tag")

    #   A       A       A
    # [ | | | | | | | . . . . . ]
    #   0                     11
    out_new = vgprPool.findFreeRange(5, 4)
    out_old = findFreeRange_oldLogic(vgprPool, 5, 4)

    assert out_new == out_old == None


def test_findFreeRange_noOverflowButPastLastSlot():
    poolSize = 12
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=False, printRP=0)

    vgprPool.add(7, 5, "tag")

    #   A       A       A
    # [ | | | | | | | . . . . . ]
    #   0                     11
    out_new = vgprPool.findFreeRange(5, 4)
    out_old = findFreeRange_oldLogic(vgprPool, 5, 4)

    assert out_new == out_old
    assert out_new == 8


def test_findFreeRange_overflowPastPoolLength():
    poolSize = 12
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=True, printRP=0)

    vgprPool.add(0, poolSize, "tag")

    #   A       A       A
    # [ . . . . . . . . . . . . ]
    #   0                     11
    out_new = vgprPool.findFreeRange(16, 4)
    out_old = findFreeRange_oldLogic(vgprPool, 16, 4)

    assert out_new == out_old == None


def test_findFreeRange_noOverflowPastPoolLength():
    poolSize = 12
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=False, printRP=0)

    vgprPool.add(0, poolSize, "tag")

    #   A       A       A
    # [ . . . . . . . . . . . . ]
    #   0                     11
    out_new = vgprPool.findFreeRange(16, 4)
    out_old = findFreeRange_oldLogic(vgprPool, 16, 4)

    assert out_new == out_old
    assert out_new == 0


def test_findFreeRange_noOverflowPastPoolLength2():
    poolSize = 12
    vgprPool = RegisterPool(poolSize, "v", defaultPreventOverflow=False, printRP=0)

    vgprPool.add(0, poolSize - 2, "tag")

    print(vgprPool.pool)
    #   A       A       A
    # [ | | | | | | | | | | . . ]
    #   0                     11
    out_new = vgprPool.findFreeRange(16, 8)
    out_old = findFreeRange_oldLogic(vgprPool, 16, 8)

    assert out_new == out_old
    assert out_new == 16


# ----------------
# Helper functions
# ----------------
def findFreeRange_oldLogic(
    regPool, size, alignment, preventOverflow=-1, wantedStatus=RegisterPool.Status.Available
):
    if preventOverflow == -1:
        preventOverflow = regPool.defaultPreventOverflow

    for i in range(len(regPool.pool) + 1):
        if i % alignment != 0:
            continue
        if regPool.isRangeAvailable(i, size, preventOverflow, wantedStatus):
            return i

    if preventOverflow:
        return None
    else:
        loc = regPool.startOfLastAvailableBlock()
        return roundUpToNearestMultiple(loc, alignment)



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

"""
A component is a piece of code that is chosen among compatible components based on the current hardware capabilities and/or kernel options.

The class hierarchy is automatically used to detect which type of component a given class belongs to. For example, all the MAC components should inherit from the MAC class.

Most components should be able to get away with defining their requirements via the class properties (e.g.):

```python
    class FMA_NonPacked(MAC):
        asmCaps = {"v_fma_f16": True,
                "v_pk_fma_f16": False}
        #archCaps = {}
        kernel = {"ProblemType": {"DataType": DataType(DataType.half),
                                "HighPrecisionAccumulate": False}}
```

Values in the dictionaries can be lambdas for more advanced logic:

```python
    class FMA_HPA_MAD_MIX(MAC):
        asmCaps = {"v_mad_mix_f32": True}
        #archCaps = {}
        kernel = {"ProblemType": {"DataType": DataType(DataType.half),
                                "HighPrecisionAccumulate": True},
                "LocalDotLayout": lambda ldl: ldl > 1
                }
```

Any more advanced logic should be implemented by overriding the matches() method.

Components are found by calling `Component.<subtype>.find(writer)` where `writer` is a `KernelWriter` object:

```python
    component = Component.MAC.find(self)
    if component:
      return component(self, m, innerUnroll)

    # No component was found, fall back to existing code...
```

With this fallback mechanism, components can be added one at a time, without disrupting existing code.

Components can be categorized in different files in the `Tensile/Components` directory.  Each file should be listed in the `__all__` member of `Tensile/Components/__init__.py`.
"""

import abc
from collections.abc import Mapping
import inspect

def PartialMatch(pattern, obj, debug=False, level=0):
    indent = "    " * level
    if debug and level == 0:
        print("")
    if hasattr(pattern, "__call__"):
        if not pattern(obj):
            if debug:
                print("{indent}call({obj}) == False".format(indent=indent, obj=obj))
            return False
    elif isinstance(pattern, Mapping) and \
         isinstance(obj, Mapping):
        for key, value in pattern.items():
            if key not in obj:
                if debug:
                    print("{indent}{key} not in object.".format(indent=indent, key=key))
                return False

            if debug:
                print("{indent} recursing into {key}".format(indent=indent, key=key))
            if not PartialMatch(value, obj[key], debug, level+1):
                return False

    elif pattern != obj:
        if debug:
            print("{indent}{pattern} != {obj}".format(indent=indent, pattern=pattern, obj=obj))
        return False

    if debug:
        print("{indent}: True".format(indent=indent))
    return True

class ComponentMeta(abc.ABCMeta):
    """
    Metaclass which auto-registers each subclass in an "implementations"
    member of its parent class, to allow for hierarchical searching.
    """
    def __init__(cls, name, bases, namespace, **kwargs):
        if inspect.isabstract(cls):
            cls.implementations = {}

        for base in bases:
            base.implementations[name] = cls
            setattr(base, name, cls)

class Component(metaclass=ComponentMeta):
    """
    Modular component which allows kernel components to be specified based on
    capability rather than based on individual architecture IDs.
    """

    @classmethod
    def matches(cls, writer, debug=False):
        if hasattr(cls, "versions"):
            if not writer.version in cls.versions:
                return False

        attrs = ["asmCaps", "archCaps", "kernel"]
        for attr in attrs:
            if hasattr(cls, attr):
                if not PartialMatch(getattr(cls, attr), getattr(writer, attr), debug):
                    return False

        return True

    @classmethod
    def findAll(cls, writer, debug=False, *args, **kwargs):
        found = []
        for name, impl in cls.implementations.items():
            if debug:
                print(name)
            if inspect.isabstract(impl):
                foundHere = impl.findAll(writer, debug, *args, **kwargs)
                if debug:
                    print(name, ": found ", foundHere)
                found += foundHere
            elif impl.matches(writer, debug, *args, **kwargs):
                if debug:
                    print(name, ": found impl")
                found.append(impl)
            else:
                if debug:
                    print(name, "mismatch!")

        return found

    @classmethod
    def find(cls, writer, debug=False, *args, **kwargs):
        found = cls.findAll(writer, debug, *args, **kwargs)

        if len(found) == 0:
            return None

        if len(found) > 1:
            raise RuntimeError("Found {} implementations for {}".format(len(found), cls.__name__))

        return found[0]()

    @classmethod
    def componentPath(cls, path=None, bases=None):
        if path is None:
            path = []

        if bases is None:
            bases = cls.__bases__

        if not isinstance(cls, str):
            className = cls.__name__
        path = [className] + path

        if cls == Component or len(bases) == 0:
            return path
        return bases[0].componentPath(path)

    @abc.abstractmethod
    def __call__(self):
        """
        Concrete subclasses must implement __call__.
        """
        pass

    def commentHeader(self):
        """
        Returns a comment which helps identify where a piece of code was generated.
        """
        return "// {}\n".format('.'.join(self.componentPath()))

class MAC(Component):
    """
    Multiply-accumulate block.
    """
    pass

class MFMA(Component):
    """
    MFMA instruction selection.
    """

    def WaitCount(self, writer):
        return 0

    pass

class Signature(Component):
    """
    Function signature block.
    """
    pass

class LocalRead(Component):
    """
    Local read block.
    """
    pass

class ShiftVectorComponents(Component):
    """
    Shift vector components block.
    """
    pass

class ComputeStoreVgprs(Component):
    """
    Compute store vgprs block.
    """
    pass

class NotLocalFullTileElements(Component):
    """
    Not local full tile elements block.
    """
    pass

class LraTileAssignment(Component):
    """
    Lra tile assignment block.
    """
    pass

class PseudoRandomGenerator(Component):
    """
    Pseudo random generator block.
    """
    pass

# Importing here allows auto-registry of components in the Components directory.
# Each file must be listed in __all__ in Components/__init__.py
# "noqa" prevents linter from complaining here.
from .Components import *  # noqa



################################################################################
#
# Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from Tensile.KernelWriterAssembly import KernelWriterAssembly

def test_occupancy():
    kw = KernelWriterAssembly("","")
    # numThreads = 256
    assert kw.getOccupancy(256,  10, 65536,   0) == 1
    assert kw.getOccupancy(256,  10, 16384, 128) == 2
    assert kw.getOccupancy(256,  65,  8192,  64) == 3
    assert kw.getOccupancy(256,  10, 65536) == 1
    assert kw.getOccupancy(256,  10, 32768) == 2
    assert kw.getOccupancy(256, 129, 32768) == 1
    assert kw.getOccupancy(256,  10, 16384) == 4
    assert kw.getOccupancy(256, 256, 32768) == 1

    # numThreads = 512
    assert kw.getOccupancy(512,  10, 65536,   0) == 1
    assert kw.getOccupancy(512,  10, 16384, 128) == 1
    assert kw.getOccupancy(512,  65,  8192,  64) == 1
    assert kw.getOccupancy(512,  10, 65536) == 1
    assert kw.getOccupancy(512,  10, 32768) == 2
    assert kw.getOccupancy(512, 129, 32768) == 0
    assert kw.getOccupancy(512,  10, 16384) == 4
    assert kw.getOccupancy(512, 256, 32768) == 0
    assert kw.getOccupancy(512,  13,  1024) == 5

    #----------- unifiedVgprRegs = True ---------------
    # numThreads = 256
    assert kw.getOccupancy(256, 256, 32768, 128, True) ==  1
    assert kw.getOccupancy(256, 128, 65536, 128, True) ==  1
    assert kw.getOccupancy(256, 256, 32768,   0, True) ==  2
    assert kw.getOccupancy(256, 128, 32768, 128, True) ==  2
    assert kw.getOccupancy(256, 128,  1024, 128, True) ==  2
    assert kw.getOccupancy(256,  97,  8192,   0, True) ==  4
    assert kw.getOccupancy(256,  65,  8192,  32, True) ==  4
    assert kw.getOccupancy(256,  64,  8192,  32, True) ==  5
    assert kw.getOccupancy(256,  65,  8192,   0, True) ==  7
    assert kw.getOccupancy(256,   0,  8192,  64, True) ==  8
    assert kw.getOccupancy(256,  24,  8192,  24, True) ==  8
    assert kw.getOccupancy(256,  49,  1024,   0, True) ==  9
    assert kw.getOccupancy(256,  24,  1024,  24, True) == 10
    assert kw.getOccupancy(256,  48,  1024,   0, True) == 10
    assert kw.getOccupancy(256,   3,  1024,  16, True) == 10

    # numThreads = 512
    assert kw.getOccupancy(512, 256, 32768, 128, True) ==  0
    assert kw.getOccupancy(512, 128, 65536, 128, True) ==  1
    assert kw.getOccupancy(512, 256, 32768,   0, True) ==  1
    assert kw.getOccupancy(512, 128, 32768, 128, True) ==  1
    assert kw.getOccupancy(512, 128,  1024, 128, True) ==  1
    assert kw.getOccupancy(512,  97,  8192,   0, True) ==  2
    assert kw.getOccupancy(512,  65,  8192,  32, True) ==  2
    assert kw.getOccupancy(512,  64,  8192,  32, True) ==  2
    assert kw.getOccupancy(512,  65,  8192,   0, True) ==  3
    assert kw.getOccupancy(512,   0,  8192,  64, True) ==  4
    assert kw.getOccupancy(512,  49,  1024,   0, True) ==  4
    assert kw.getOccupancy(512,  24,  1024,  24, True) ==  5
    assert kw.getOccupancy(512,  48,  1024,   0, True) ==  5
    assert kw.getOccupancy(512,   3,  1024,  16, True) ==  5

def test_max_regs():
    kw = KernelWriterAssembly("","")
    assert kw.getMaxRegsForOccupancy(256,  10, 65536,   0) == 256
    assert kw.getMaxRegsForOccupancy(256,  10, 16384, 128) == 128
    assert kw.getMaxRegsForOccupancy(256,  65,  8192,  64) == 84
    assert kw.getMaxRegsForOccupancy(256,  10, 65536) == 256
    assert kw.getMaxRegsForOccupancy(256,  10, 32768) == 128
    assert kw.getMaxRegsForOccupancy(256, 129, 32768) == 256
    assert kw.getMaxRegsForOccupancy(256,  10, 16384) == 64
    assert kw.getMaxRegsForOccupancy(256, 256, 32768) == 256

    # numThreads = 512
    assert kw.getMaxRegsForOccupancy(512,  10, 65536,   0) == 128
    assert kw.getMaxRegsForOccupancy(512,  10, 16384, 128) == 128
    assert kw.getMaxRegsForOccupancy(512,  65,  8192,  64) == 128
    assert kw.getMaxRegsForOccupancy(512,  10, 65536) == 128
    assert kw.getMaxRegsForOccupancy(512,  10, 32768) ==  64
    assert kw.getMaxRegsForOccupancy(512, 129, 32768) == 129
    assert kw.getMaxRegsForOccupancy(512,  10, 16384) ==  32
    assert kw.getMaxRegsForOccupancy(512, 256, 32768) == 256
    assert kw.getMaxRegsForOccupancy(512,  13,  1024) ==  24

    #----------- unifiedVgprRegs = True ---------------
    # numThreads = 256
    assert kw.getMaxRegsForOccupancy(256, 256, 32768, 128, True) ==  256
    assert kw.getMaxRegsForOccupancy(256, 128, 65536, 128, True) ==  256
    assert kw.getMaxRegsForOccupancy(256, 256, 32768,   0, True) ==  256
    assert kw.getMaxRegsForOccupancy(256, 128, 32768, 128, True) ==  128
    assert kw.getMaxRegsForOccupancy(256, 128,  1024, 128, True) ==  128
    assert kw.getMaxRegsForOccupancy(256,  97,  8192,   0, True) ==  128
    assert kw.getMaxRegsForOccupancy(256,  65,  8192,  32, True) ==   96
    assert kw.getMaxRegsForOccupancy(256,  64,  8192,  32, True) ==   64
    assert kw.getMaxRegsForOccupancy(256,  65,  8192,   0, True) ==   72
    assert kw.getMaxRegsForOccupancy(256,   0,  8192,  64, True) ==    0
    assert kw.getMaxRegsForOccupancy(256,  24,  8192,  24, True) ==   40
    assert kw.getMaxRegsForOccupancy(256,  49,  1024,   0, True) ==   56
    assert kw.getMaxRegsForOccupancy(256,  24,  1024,  24, True) ==   24
    assert kw.getMaxRegsForOccupancy(256,  48,  1024,   0, True) ==   48
    assert kw.getMaxRegsForOccupancy(256,   3,  1024,  16, True) ==   32

    # numThreads = 512
    assert kw.getMaxRegsForOccupancy(512, 256, 32768, 128, True) ==  256
    assert kw.getMaxRegsForOccupancy(512, 128, 65536, 128, True) ==  128
    assert kw.getMaxRegsForOccupancy(512, 256, 32768,   0, True) ==  256
    assert kw.getMaxRegsForOccupancy(512, 128, 32768, 128, True) ==  128
    assert kw.getMaxRegsForOccupancy(512, 128,  1024, 128, True) ==  128
    assert kw.getMaxRegsForOccupancy(512,  97,  8192,   0, True) ==  128
    assert kw.getMaxRegsForOccupancy(512,  65,  8192,  32, True) ==   96
    assert kw.getMaxRegsForOccupancy(512,  64,  8192,  32, True) ==   96
    assert kw.getMaxRegsForOccupancy(512,  65,  8192,   0, True) ==   80
    assert kw.getMaxRegsForOccupancy(512,   0,  8192,  64, True) ==    0
    assert kw.getMaxRegsForOccupancy(512,  49,  1024,   0, True) ==   64
    assert kw.getMaxRegsForOccupancy(512,  24,  1024,  24, True) ==   24
    assert kw.getMaxRegsForOccupancy(512,  48,  1024,   0, True) ==   48
    assert kw.getMaxRegsForOccupancy(512,   3,  1024,  16, True) ==   32

# test_occupancy()
# test_max_regs()



################################################################################
#
# Copyright (C) 2019-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from . import Properties
from . import Common
import copy

class HardwarePredicate(Properties.Predicate):
    @classmethod
    def FromISA(cls, isa):
        gfxArch = Common.gfxName(isa)
        return cls("AMDGPU", value=cls("Processor", value=gfxArch))

    @classmethod
    def FromHardware(cls, isa, cuCount=None, isAPU=None):
        gfxArch = Common.gfxName(isa)
        if cuCount == None and isAPU == None:
            return cls("AMDGPU", value=cls("Processor", value=gfxArch))
        elif cuCount == None:
            return cls("AMDGPU", value=cls.And([cls("Processor", value=gfxArch),
                                                cls("IsAPU", value=isAPU)]))
        elif isAPU == None:
            return cls("AMDGPU", value=cls.And([cls("Processor", value=gfxArch),
                                                cls("CUCount", value=cuCount)]))
        else:
            return cls("AMDGPU", value=cls.And([cls("Processor", value=gfxArch),
                                                cls("CUCount", value=cuCount),
                                                cls("IsAPU", value=isAPU)]))

    def __lt__(self, other):
        # Use superclass logic for TruePreds
        if other.tag == 'TruePred' or self.tag == 'TruePred':
            return super().__lt__(other)

        # Compute unit counts or APU/XPU versions are embedded as 'And' with
        # 'Processor', 'CUCount', and 'IsAPU' as children
        if self.value.tag == 'And':
            myAndPred = self.value
            myProcPred = next(iter(x for x in myAndPred.value if x.tag == "Processor"), None)
            myCUPred = next(iter(x for x in myAndPred.value if x.tag == "CUCount"), None)
            myCUCount = myCUPred.value if myCUPred != None else 0
            myIsAPUPred = next(iter(x for x in myAndPred.value if x.tag == "IsAPU"), None)
            myIsAPU = myIsAPUPred.value if myIsAPUPred != None else -1
        else:
            myProcPred = self.value
            myCUCount = 0
            myIsAPU = -1

        if other.value.tag == 'And':
            otherAndPred = other.value
            otherProcPred = next(iter(x for x in otherAndPred.value if x.tag == "Processor"), None)
            otherCUPred = next(iter(x for x in otherAndPred.value if x.tag == "CUCount"), None)
            otherCUCount = otherCUPred.value if otherCUPred != None else 0
            otherIsAPUPred = next(iter(x for x in otherAndPred.value if x.tag == "IsAPU"), None)
            otherIsAPU = otherIsAPUPred.value if otherIsAPUPred != None else -1
        else:
            otherProcPred = other.value
            otherCUCount = 0
            otherIsAPU = -1

        # If APU properties are the same, then check CU count or architecture
        if myIsAPU == otherIsAPU:
            # If CU properties are empty, then compare processor predicates
            if myCUCount == otherCUCount == 0:
                # Make sure that we have valid processor preds
                assert myProcPred != None and otherProcPred != None, "Missing processor predicate"
                assert myProcPred.tag == otherProcPred.tag == "Processor", "Invalid processor predicate"

                # Downgrade to base class so that we don't recurse
                myProcPredCopy = copy.deepcopy(myProcPred)
                otherProcPredCopy = copy.deepcopy(otherProcPred)
                myProcPredCopy.__class__ = otherProcPredCopy.__class__ = Properties.Predicate
                return myProcPredCopy < otherProcPredCopy

            # Higher priority given to higher CU count
            return myCUCount > otherCUCount
        # APU sorted before XPU, and XPU sorted before generic
        return myIsAPU > otherIsAPU



################################################################################
#
# Copyright (C) 2021-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import os

class ScriptHelper(object):
    """
    Helper class to facilitate formatting when
    writing bash scripts.
    """

    @classmethod
    def genLine(cls, text):
        return str("{0}\n").format(text)

    @classmethod
    def genComment(cls, text):
        return cls.genLine(str("#") + str(text))

    @classmethod
    def genEcho(cls, text):
        return cls.genLine(str("echo \"{0}\";").format(str(text)))

    @classmethod
    def makeExecutable(cls, path):
        mode = os.stat(path).st_mode
        mode |= (mode & 0o444) >> 2    # copy R bits to X
        os.chmod(path, mode)

class ScriptWriter(object):
    """
    A script writing utility class that holds a string
    buffer and wraps writing to the buffer.
    Has factory functions to build scripts with regular
    workflow interface.
    """
    def __init__(self):
        self._buffer = ""

    def writeLine(self, text):
        self._buffer += ScriptHelper.genLine(text)

    def writeComment(self, text):
        self._buffer += ScriptHelper.genComment(text)

    def writeEcho(self, text):
        self._buffer += ScriptHelper.genEcho(text)

    def dumpToFile(self, fileName, outputDir = "."):
        scriptPath = os.path.join(outputDir, fileName)
        with open(scriptPath, "w") as f:
            f.write(self._buffer)
        ScriptHelper.makeExecutable(scriptPath)

    def __str__(self):
        return self._buffer

    @classmethod
    def writeScript(bufferClass, writerClass, fileName, outputDir = "."):
        buffer = bufferClass()
        writerClass.writeHeader(buffer)
        writerClass.writeBody(buffer)
        writerClass.writeEpilogue(buffer)
        buffer.dumpToFile(fileName, outputDir)

    @classmethod
    def previewScript(bufferClass, writerClass):
        buffer = bufferClass()
        writerClass.writeHeader(buffer)
        writerClass.writeBody(buffer)
        writerClass.writeEpilogue(buffer)
        print(buffer)

    @classmethod
    def writeBenchmarkNodeScript(cls, fileName, outputDir="."):
        cls.writeScript(BenchmarkNodeWriter, fileName, outputDir)

    @classmethod
    def writeBenchmarkTaskScript(cls, fileName, outputDir="."):
        cls.writeScript(BenchmarkTaskWriter, fileName, outputDir)

    @classmethod
    def writeBenchmarkJobScript(cls, fileName, outputDir="."):
        cls.writeScript(BenchmarkJobWriter, fileName, outputDir)


class BenchmarkNodeWriter(object):
    """
    Node host run script.
    Creates an instance of the docker image and
    mounts the logs, results and tasks directories
    and finally invokes the benchmark on the
    supplied yaml.
    """
    @staticmethod
    def __initHeader(sBuffer):
        sBuffer.writeComment("!/bin/bash")
        sBuffer.writeComment("Benchmark node script")
        sBuffer.writeComment("Actual host task to be run by a SLURM worker node")
        sBuffer.writeLine("")

    @staticmethod
    def __initializeFuncs(sBuffer):
        sBuffer.writeComment("Funcs")
        sBuffer.writeLine("usage() { echo \"Usage: $0 [-i <path_to_docker_image>] [-l <log_dir>] [-r <result_dir>] [-t <task_dir>]\" 1>&2; exit 1; }")
        sBuffer.writeLine("failAndExit() { echo \"FAILED: $1\" 1>&2; exit 1; }")
        sBuffer.writeLine("")

    @staticmethod
    def __parseArgs(sBuffer):
        sBuffer.writeComment("Parse arguments")
        sBuffer.writeLine("while getopts i:l:r:t: flag")
        sBuffer.writeLine("do")
        sBuffer.writeLine("    case \"${flag}\" in")
        sBuffer.writeLine("        i) dockerImagePath=${OPTARG};;")
        sBuffer.writeLine("        l) logDir=${OPTARG};;")
        sBuffer.writeLine("        r) resultDir=${OPTARG};;")
        sBuffer.writeLine("        t) taskDir=${OPTARG};;")
        sBuffer.writeLine("        *) usage;;")
        sBuffer.writeLine("    esac")
        sBuffer.writeLine("done")
        sBuffer.writeLine("[ -z \"${dockerImagePath}\" ] || [ -z \"${logDir}\" ] || [ -z \"${resultDir}\" ]  || [ -z \"${taskDir}\" ] && usage;")
        sBuffer.writeLine("[ -d \"${logDir}\" ] && [ -d \"${resultDir}\" ] && [ -d \"${taskDir}\" ] || failAndExit \"Directory not found\";")
        sBuffer.writeLine("[ -f \"${dockerImagePath}\" ] || failAndExit \"Docker image not found: $dockerImagePath\";")
        sBuffer.writeEcho("Docker image path: $dockerImagePath")
        sBuffer.writeEcho("Path to log: $logDir")
        sBuffer.writeEcho("Path to task: $taskDir")
        sBuffer.writeEcho("Path to result: $resultDir")
        sBuffer.writeLine("")

    @staticmethod
    def __loadDockerImage(sBuffer):
        sBuffer.writeComment("Import docker image")
        sBuffer.writeComment("Successful output is \"Loaded image: imageName:TAG\"")
        sBuffer.writeEcho("Loading docker image...")
        sBuffer.writeLine("dockerLoadOutput=`docker load < $dockerImagePath || failAndExit \"Importing docker image\"`")
        sBuffer.writeLine("")
        sBuffer.writeEcho("Docker Load Result: $dockerLoadOutput")
        sBuffer.writeLine("[ -z \"${dockerLoadOutput}\" ] && failAndExit \"Loading docker image $dockerImagePath\"")
        sBuffer.writeLine("")
        sBuffer.writeComment("Split the output on ':' or spaces")
        sBuffer.writeComment("Capture the docker image name and tag")
        sBuffer.writeLine("outputSplit=(${dockerLoadOutput//:/ })")
        sBuffer.writeLine("dockerName=${outputSplit[2]}")
        sBuffer.writeLine("dockerTag=${outputSplit[3]}")
        sBuffer.writeLine("")
        sBuffer.writeComment("Find docker image record with correct name and tag")
        sBuffer.writeLine("dockerImageQuery=`docker images | grep -i \"$dockerName\" | grep -i \"$dockerTag\"`")
        sBuffer.writeLine("[ -z \"${dockerImageQuery}\" ] && failAndExit \"Finding docker image record for $dockerName:$dockerTag\"")
        sBuffer.writeLine("")
        sBuffer.writeComment("Get docker image ID")
        sBuffer.writeLine("outputSplit=(${dockerImageQuery//   / })")
        sBuffer.writeLine("dockerImageId=${outputSplit[2]}")
        sBuffer.writeLine("[ -z \"${dockerImageId}\" ] && failAndExit \"Finding docker image id for $dockerName:$dockerTag\"")
        sBuffer.writeEcho("Success: Loaded Image ID: ${dockerImageId}")
        sBuffer.writeLine("")

    @staticmethod
    def __runTask(sBuffer):
        sBuffer.writeComment("Check the task dir exists")
        sBuffer.writeLine("")
        sBuffer.writeComment("Run container with mounted taskDir")
        sBuffer.writeEcho("Running container... (this might take some time):")
        sBuffer.writeLine("runCmd=\"docker run --rm --network=host --device=/dev/kfd --device=/dev/dri --group-add video --cap-add=SYS_PTRACE --security-opt seccomp=unconfined -v \"$taskDir\":\"/TaskDir\" -v \"$resultDir\":\"/ResultDir\" -v \"$logDir\":\"/LogDir\" $dockerImageId\"")
        sBuffer.writeEcho("$runCmd")
        sBuffer.writeLine("$runCmd")
        sBuffer.writeLine("returnCode=$?")

    @staticmethod
    def __exit(sBuffer):
        sBuffer.writeEcho("Done!")
        sBuffer.writeLine("exit $returnCode")

    @classmethod
    def writeHeader(cls, sBuffer):
        cls.__initHeader(sBuffer)

    @classmethod
    def writeBody(cls, sBuffer):
        cls.__initializeFuncs(sBuffer)
        cls.__parseArgs(sBuffer)
        cls.__loadDockerImage(sBuffer)
        cls.__runTask(sBuffer)

    @classmethod
    def writeEpilogue(cls, sBuffer):
        cls.__exit(sBuffer)

class BenchmarkTaskWriter(object):
    """
    SLURM specific script.
    Invokes the 'srun' command to queue a task.
    SLURM environment variables tell us which
    indexed task to run as part of the array.
    The final task submitted is the node script
    which will run on one of the hosts.
    """
    @staticmethod
    def __initHeader(sBuffer):
        sBuffer.writeComment("!/bin/bash")
        sBuffer.writeComment("Benchmark task enqueue script")
        sBuffer.writeComment("This script will configure and commit a single task to the cluster.")
        sBuffer.writeComment("Intended to be the run target of SLURM SBATCH multiplexer")
        sBuffer.writeLine("")

    @staticmethod
    def __initializeFuncs(sBuffer):
        sBuffer.writeComment("Funcs")
        sBuffer.writeLine("usage() { echo \"Usage: $0 [-i <image_dir>] [-l <logs_dir>] [-r <results_dir>] [-t <tasks_dir>] \" 1>&2; exit 1; }")
        sBuffer.writeLine("failAndExit() { echo \"FAILED: $1\" 1>&2; exit 1; }")
        sBuffer.writeLine("")

    @staticmethod
    def __parseArgs(sBuffer):
        sBuffer.writeComment("Parse arguments")
        sBuffer.writeLine("while getopts i:l:r:t: flag")
        sBuffer.writeLine("do")
        sBuffer.writeLine("    case \"${flag}\" in")
        sBuffer.writeLine("        i) imageDir=${OPTARG};; ")
        sBuffer.writeLine("        l) logsDir=${OPTARG};;")
        sBuffer.writeLine("        r) resultsDir=${OPTARG};;")
        sBuffer.writeLine("        t) tasksDir=${OPTARG};;")
        sBuffer.writeLine("        *) usage;;")
        sBuffer.writeLine("    esac")
        sBuffer.writeLine("done")
        sBuffer.writeLine("[ -z \"${imageDir}\" ] || [ -z \"${logsDir}\" ] || [ -z \"${resultsDir}\" ] || [ -z \"${tasksDir}\" ] && usage;")
        sBuffer.writeLine("[ -d \"${imageDir}\" ] && [ -d \"${logsDir}\" ] && [ -d \"${resultsDir}\" ] && [ -d \"${tasksDir}\" ] || failAndExit \"Directory not found\";")
        sBuffer.writeEcho("Image dir: $imageDir")
        sBuffer.writeEcho("Logs dir: $logsDir")
        sBuffer.writeEcho("Tasks dir: $tasksDir")
        sBuffer.writeEcho("Results dir: $resultsDir")
        sBuffer.writeLine("")

    @staticmethod
    def __configureTask(sBuffer):
        sBuffer.writeComment("Setup task parameters")
        sBuffer.writeLine("pushd $tasksDir")
        sBuffer.writeLine("subDirs=(*)")
        sBuffer.writeLine("taskName=${subDirs[$SLURM_ARRAY_TASK_ID]}")
        sBuffer.writeLine("taskDir=$tasksDir/$taskName")
        sBuffer.writeLine("taskResultDir=$resultsDir/$taskName")
        sBuffer.writeLine("mkdir -p $taskResultDir")
        sBuffer.writeLine("taskLogDir=$logsDir/$taskName")
        sBuffer.writeLine("mkdir -p $taskLogDir")
        sBuffer.writeLine("task=($taskDir/*.sh)")
        sBuffer.writeLine("image=($imageDir/*.tar.gz)")
        sBuffer.writeEcho("Task ID: $SLURM_ARRAY_TASK_ID / $SLURM_TASK_MAX")
        sBuffer.writeEcho("Task Name: $taskName")
        sBuffer.writeEcho("Host Name: `hostname`")
        sBuffer.writeLine("")

    @staticmethod
    def __submitTask(sBuffer):
        sBuffer.writeComment("Enqueue node task")
        sBuffer.writeLine("srun -N 1 \"$task\" -i \"$image\" -l \"$taskLogDir\" -r \"$taskResultDir\" -t \"$taskDir\"")
        sBuffer.writeLine("returnCode=$?")

    @staticmethod
    def __exit(sBuffer):
        sBuffer.writeEcho("Done!")
        sBuffer.writeLine("exit $returnCode")

    @classmethod
    def writeHeader(cls, sBuffer):
        cls.__initHeader(sBuffer)

    @classmethod
    def writeBody(cls, sBuffer):
        cls.__initializeFuncs(sBuffer)
        cls.__parseArgs(sBuffer)
        cls.__configureTask(sBuffer)
        cls.__submitTask(sBuffer)

    @classmethod
    def writeEpilogue(cls, sBuffer):
        cls.__exit(sBuffer)

class BenchmarkJobWriter(object):
    """
    SLURM specific script.
    Invokes the 'sbatch' and allocates resources.
    Will create an array of N tasks that will each
    run one instance of the task script.
    N = number of sub-folders inside task folder
    Assumes one node per task.
    Will wait until the entire batch is complete
    before exiting.
    NOTE: the wait feature requires that all task
    scripts have a return code, or will hang forever!
    """
    @staticmethod
    def __initHeader(sBuffer):
        sBuffer.writeComment("!/bin/bash")
        sBuffer.writeComment("Benchmark batch job script")
        sBuffer.writeComment("This script will configure and submit a batch job to the cluster")
        sBuffer.writeComment("This is the SLURM cluster benchmark entry-point")
        sBuffer.writeLine("")

    @staticmethod
    def __initializeFuncs(sBuffer):
        sBuffer.writeComment("Funcs")
        sBuffer.writeLine("usage() { echo \"Usage: $0 [-i <image_dir>] [-l <logs_dir>] [-r <results_dir>] [-s <path_to_task_script>] [-t <tasks_dir>] \" 1>&2; exit 1; }")
        sBuffer.writeLine("failAndExit() { echo \"FAILED: $1\" 1>&2; exit 1; }")
        sBuffer.writeLine("")

    @staticmethod
    def __parseArgs(sBuffer):
        sBuffer.writeComment("Parse arguments")
        sBuffer.writeLine("while getopts i:l:r:s:t: flag")
        sBuffer.writeLine("do")
        sBuffer.writeLine("    case \"${flag}\" in")
        sBuffer.writeLine("        i) imageDir=${OPTARG};;")
        sBuffer.writeLine("        l) logsDir=${OPTARG};;")
        sBuffer.writeLine("        r) resultsDir=${OPTARG};;")
        sBuffer.writeLine("        s) taskScript=${OPTARG};;")
        sBuffer.writeLine("        t) tasksDir=${OPTARG};;")
        sBuffer.writeLine("        *) usage;;")
        sBuffer.writeLine("    esac")
        sBuffer.writeLine("done")
        sBuffer.writeLine("[ -z \"${imageDir}\" ] || [ -z \"${logsDir}\" ] || [ -z \"${resultsDir}\" ] || [ -z \"${tasksDir}\" ]  && usage;")
        sBuffer.writeLine("[ -d \"${imageDir}\" ] && [ -d \"${logsDir}\" ] && [ -d \"${resultsDir}\" ] && [ -d \"${tasksDir}\" ] || failAndExit \"Directory not found\";")
        sBuffer.writeLine("[ -f \"${taskScript}\" ] || failAndExit \"Task script not found\";")
        sBuffer.writeEcho("Image dir: $imageDir")
        sBuffer.writeEcho("Logs dir: $logsDir")
        sBuffer.writeEcho("Tasks dir: $tasksDir")
        sBuffer.writeEcho("Results dir: $resultsDir")
        sBuffer.writeEcho("Enqueue script: $taskScript")
        sBuffer.writeLine("")

    @staticmethod
    def __configureJob(sBuffer):
        sBuffer.writeComment("Setup local environment")
        sBuffer.writeComment("Get in the directory where this script lives")
        sBuffer.writeLine("pushd $( dirname `greadlink -f ${BASH_SOURCE[0]} || readlink -f ${BASH_SOURCE[0]}` )")
        sBuffer.writeComment("Enumerate the tasks and set up batch params based on task count")
        sBuffer.writeLine("subDirs=($tasksDir/*)")
        sBuffer.writeLine("let arraySize=${#subDirs[@]}")
        sBuffer.writeLine("[ -z \"${arraySize}\" ] && failAndExit \"No task count\";")
        sBuffer.writeLine("let arrayStart=0")
        sBuffer.writeLine("let arrayEnd=$arraySize-1")
        sBuffer.writeComment("Create logs directory for SLURM")
        sBuffer.writeLine("slurmLogsDir=$logsDir/SLURM")
        sBuffer.writeLine("mkdir -p $slurmLogsDir")
        sBuffer.writeLine("")

    @staticmethod
    def __submitJob(sBuffer):
        sBuffer.writeComment("Invoke batch request to SLURM")
        sBuffer.writeComment("Will wait until the task completes before exiting")
        sBuffer.writeLine("runCmd=\"sbatch --nodes=1 --array=$arrayStart-$arrayEnd -o $slurmLogsDir/slurm-%A_%a.out --wait $taskScript -i $imageDir -l $logsDir -r $resultsDir -t $tasksDir\"")
        sBuffer.writeEcho("Submitting batch job to SLURM. This may take a while...")
        sBuffer.writeEcho("$runCmd")
        sBuffer.writeLine("$runCmd")
        sBuffer.writeLine("returnCode=$?")
        sBuffer.writeLine("popd")

    @staticmethod
    def __exit(sBuffer):
        sBuffer.writeEcho("Done!")
        sBuffer.writeLine("exit $returnCode")

    @classmethod
    def writeHeader(cls, sBuffer):
        cls.__initHeader(sBuffer)

    @classmethod
    def writeBody(cls, sBuffer):
        cls.__initializeFuncs(sBuffer)
        cls.__parseArgs(sBuffer)
        cls.__configureJob(sBuffer)
        cls.__submitJob(sBuffer)

    @classmethod
    def writeEpilogue(cls, sBuffer):
        cls.__exit(sBuffer)



################################################################################
#
# Copyright (C) 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import itertools
import os
import re
import shlex
import shutil
import subprocess
from pathlib import Path
from typing import List, Union

from ..Common import (
    IsaVersion,
    ParallelMap,
    ensurePath,
    globalParameters,
    splitArchs,
    tPrint,
)


def _computeSourceCodeObjectPath(
    target: str, base: str, buildPath: Union[Path, str], arch
) -> Union[Path, None]:
    """Generates a code object file path using the target, base, and build path.

    Args:
        target: The target triple.
        base: The base name for the output file (name without extension).
        buildPath: The build directory path.

    Returns:
        Path to the code object file if a name can be computed, otherwise None, which
        indicates that the code object file should not be generated.
    """
    coPath = None
    buildPath = Path(buildPath)
    if "TensileLibrary" in base and "fallback" in base:
        coPath = buildPath / "{0}_{1}.hsaco.raw".format(base, arch)
    elif "TensileLibrary" in base:
        variant = [t for t in ["", "xnack-", "xnack+"] if t in target][-1]
        baseVariant = base + "-" + variant if variant else base
        if arch in baseVariant:
            coPath = buildPath / (baseVariant + ".hsaco.raw")
    else:
        coPath = buildPath / "{0}.so-000-{1}.hsaco.raw".format(base, arch)
    return coPath


def _compileSourceObjectFile(
    cmdlineArchs: List[str],
    cxxCompiler: str,
    cxxSrcPath: str,
    objDestPath: str,
    outputPath: str,
    compilerVer: IsaVersion,
):
    """Compiles a HIP source file (.cpp) into an object file (.o).

    Args:
        cmdlineArchs: List of architectures for offloading.
        cxxCompiler: The C++ compiler to use.
        cxxSrcPath: The path to the kernel source file.
        objDestPath: The build directory path.
        outputPath: The output directory path.
        compilerVer: The compiler version.

    Raises:
        subprocess.CalledProcessError: If the compilation command fails.
    """
    archFlags = ["--offload-arch=" + arch for arch in cmdlineArchs]

    hipFlags = ["-D__HIP_HCC_COMPAT_MODE__=1"]
    # TODO(@tensile-infra): hipcc is deprecated and should be removed for ROCm 6.5
    hipFlags += (
        ["--genco"]
        if os.path.basename(cxxCompiler) == "hipcc"
        else ["--cuda-device-only", "-x", "hip", "-O3"]
    )
    hipFlags += ["-I", outputPath]

    if len(compilerVer) >= 2 and (
        compilerVer[0] > 5 or (compilerVer[0] == 5 and compilerVer[1] > 2)
    ):
        hipFlags += ["-Xoffload-linker", "--build-id"]

    launcher = shlex.split(os.environ.get("Tensile_CXX_COMPILER_LAUNCHER", ""))

    if os.name == "nt":
        hipFlags += [
            "-std=c++14",
            "-fms-extensions",
            "-fms-compatibility",
            "-fPIC",
            "-Wno-deprecated-declarations",
        ]

    args = (
        launcher + [cxxCompiler] + hipFlags + archFlags + [str(cxxSrcPath), "-c", "-o", objDestPath]
    )

    tPrint(2, f"Compile source object file command: {args}")
    try:
        out = subprocess.check_output(args, stderr=subprocess.STDOUT)
        tPrint(2, f"Output: {out}" if out else "")
    except subprocess.CalledProcessError as err:
        raise RuntimeError(
            f"Error compiling source object file: {err.output}\nFailed command: {' '.join(args)}"
        )


def _listTargetTriples(bundler: str, objFile: str):
    """Invokes the bundler to list target triples in an object file.

    Args:
        bundler: The path to the bundler executable, typically ``clang-offload-bundler``.
        objFile: String-like path to the object file.
    """
    args = [bundler, "--type=o", f"--input={objFile}", "-list"]
    try:
        listing = subprocess.check_output(args, stderr=subprocess.STDOUT).decode().split("\n")
    except subprocess.CalledProcessError as err:
        raise RuntimeError(
            f"Error listing target triples in object files: {err.output}\nFailed command: {' '.join(args)}"
        )
    return listing


def _unbundleSourceCodeObjects(bundler: str, target: str, infile: str, outfileRaw: str):
    """Unbundles code object files from the provided target triple.

    Args:
        bundler (str): The path to the Clang Offload Bundler executable.
        target (str): The target architecture string.
        inflag (str): The input flag for the bundler.
        infile (str): The input file path.
        outflag (str): The output flag for the bundler.
        outfileRaw (str): The output raw file path.

    Raises:
        RuntimeError: If unbundling the source code object file fails.
    """
    args = [
        bundler,
        "--type=o",
        f"--targets={target}",
        f"--input={infile}",
        f"--output={outfileRaw}",
        "--unbundle",
    ]

    tPrint(2, "Unbundling source code object file: " + " ".join(args))
    try:
        out = subprocess.check_output(args, stderr=subprocess.STDOUT)
        tPrint(2, f"Output: {out}" if out else "")
    except subprocess.CalledProcessError as err:
        raise RuntimeError(
            f"Error unbundling source code object file: {err.output}\nFailed command: {' '.join(args)}"
        )


def _buildSourceCodeObjectFile(
    cxxCompiler: str,
    outputPath: Union[Path, str],
    kernelPath: Union[Path, str],
    removeTemporaries: bool,
) -> List[str]:

    buildPath = Path(ensurePath(os.path.join(globalParameters["WorkingPath"], "code_object_tmp")))
    destPath = Path(ensurePath(os.path.join(outputPath, "library")))
    kernelPath = Path(kernelPath)

    objectFilename = kernelPath.stem + ".o"
    coPathsRaw = []
    coPaths = []

    if "CmakeCxxCompiler" in globalParameters and globalParameters["CmakeCxxCompiler"] is not None:
        os.environ["CMAKE_CXX_COMPILER"] = globalParameters["CmakeCxxCompiler"]

    _, cmdlineArchs = splitArchs()

    # Add build-id for builds with rocm 5.3+
    compilerVer = globalParameters["HipClangVersion"].split(".")[:2]
    compilerVer = [int(c) for c in compilerVer]

    objPath = str(buildPath / objectFilename)
    _compileSourceObjectFile(
        cmdlineArchs, cxxCompiler, str(kernelPath), objPath, str(outputPath), compilerVer
    )

    bundler = globalParameters["ClangOffloadBundlerPath"]
    if not bundler:
        raise RuntimeError(
            "No bundler found; set TENSILE_ROCM_OFFLOAD_BUNDLER_PATH to point to clang-offload-bundler"
        )

    for target in _listTargetTriples(bundler, objPath):
        match = re.search("gfx.*$", target)
        if match:
            arch = re.sub(":", "-", match.group())
            coPathRaw = _computeSourceCodeObjectPath(target, kernelPath.stem, buildPath, arch)
            if not coPathRaw:
                continue
            _unbundleSourceCodeObjects(bundler, target, objPath, str(coPathRaw))

            coPath = str(destPath / coPathRaw.stem)
            coPaths.append(coPath)
            coPathsRaw.append(coPathRaw)

    for src, dst in zip(coPathsRaw, coPaths):
        shutil.copy(src, dst)
    if removeTemporaries:
        for file in coPathsRaw:
            os.remove(file)

    return coPaths


def buildSourceCodeObjectFiles(
    cxxCompiler: str, kernelFiles: List[Path], outputPath: Path, removeTemporaries: bool
) -> List[str]:
    """Compiles HIP source code files into code object files.

    Args:
        cxxCompiler: The C++ compiler to use.
        kernelFiles: List of paths to the kernel source files.
        outputPath: The output directory path where code objects will be placed.
        removeTemporaries: Whether to clean up temporary files.

    Returns:
        List of paths to the created code objects.
    """

    args = zip(
        itertools.repeat(cxxCompiler),
        itertools.repeat(outputPath),
        kernelFiles,
        itertools.repeat(removeTemporaries),
    )
    coFiles = ParallelMap(_buildSourceCodeObjectFile, args, "Compiling source kernels")

    return itertools.chain.from_iterable(coFiles)
